# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test case for module Memory
"""

import os
import re
from unittest import mock

from tests import common
from bleachbit import logger
from bleachbit.FileUtilities import exe_exists
from bleachbit.Memory import (
    count_swap_linux,
    disable_swap_linux,
    enable_swap_linux,
    get_proc_swaps,
    get_swap_size_linux,
    get_swap_uuid,
    make_self_oom_target_linux,
    parse_swapoff,
    physical_free, physical_free_darwin,
    report_free,
    wipe_memory,
    wipe_swap_linux)


class MemoryTestCase(common.BleachbitTestCase):
    """Test case for module Memory"""

    @common.skipIfWindows
    def test_get_proc_swaps(self):
        """Test for method get_proc_swaps"""
        if not exe_exists('swapon'):
            self.skipTest('swapon not found')
        ret = get_proc_swaps()
        self.assertGreater(len(ret), 10)
        if not re.search(r'Filename\s+Type\s+Size', ret):
            raise RuntimeError(
                "Unexpected first line in swap summary '%s'" % ret)

    @common.skipIfWindows
    def test_make_self_oom_target_linux(self):
        """Test for method make_self_oom_target_linux"""

        # preserve
        euid = os.geteuid()

        # Minimally test there is no traceback
        make_self_oom_target_linux()

        # restore
        os.seteuid(euid)

    @common.skipIfWindows
    def test_count_linux_swap(self):
        """Test for method count_linux_swap"""
        n_swaps = count_swap_linux()
        self.assertIsInteger(n_swaps)
        self.assertTrue(0 <= n_swaps < 10)

    @common.skipIfWindows
    def test_physical_free_darwin(self):
        self.assertEqual(physical_free_darwin(lambda:
                                              """Mach Virtual Memory Statistics: (page size of 4096 bytes)
Pages free:                              836891.
Pages active:                            588004.
Pages inactive:                           16985.
Pages speculative:                        89776.
Pages throttled:                              0.
Pages wired down:                        468097.
Pages purgeable:                          58313.
"Translation faults":                3109985921.
Pages copy-on-write:                   25209334.
Pages zero filled:                    537180873.
Pages reactivated:                    132264973.
Pages purged:                          11567935.
File-backed pages:                       184609.
Anonymous pages:                         510156.
Pages stored in compressor:              784977.
Pages occupied by compressor:             96724.
Decompressions:                        66048421.
Compressions:                          90076786.
Pageins:                              758631430.
Pageouts:                              30477017.
Swapins:                               19424481.
Swapouts:                              20258188.
"""), 3427905536)
        self.assertRaises(RuntimeError, physical_free_darwin,
                          lambda: "Invalid header")

    @common.skipIfWindows
    def test_physical_free(self):
        """Test for method physical_free"""
        ret = physical_free()
        self.assertIsInteger(
            ret, 'physical_free() returns variable type %s' % type(ret))
        self.assertGreater(physical_free(), 0)
        report_free()

    @common.skipIfWindows
    def test_get_swap_size_linux(self):
        """Test for get_swap_size_linux()"""
        if not exe_exists('swapon'):
            self.skipTest('swapon not found')
        with open('/proc/swaps', encoding='utf-8') as f:
            swapdev = f.read().split('\n')[1].split(' ')[0]
        if 0 == len(swapdev):
            self.skipTest('no active swap device detected')
        size = get_swap_size_linux(swapdev)
        self.assertIsInteger(size)
        self.assertGreater(size, 1024 ** 2)
        logger.debug("size of swap '%s': %d B (%d MB)",
                     swapdev, size, size / (1024 ** 2))
        with open('/proc/swaps', encoding='utf-8') as f:
            proc_swaps = f.read()
        size2 = get_swap_size_linux(swapdev, proc_swaps)
        self.assertEqual(size, size2)

    @common.skipIfWindows
    def test_get_swap_uuid(self):
        """Test for method get_swap_uuid"""
        if not exe_exists('blkid'):
            self.skipTest('blkid not found')
        self.assertEqual(get_swap_uuid('/dev/doesnotexist'), None)

    def test_parse_swapoff(self):
        """Test for method parse_swapoff"""
        tests = (
            # Ubuntu 15.10 has format "swapoff /dev/sda3"
            ('swapoff /dev/sda3', '/dev/sda3'),
            ('swapoff for /dev/sda6', '/dev/sda6'),
            ('swapoff on /dev/mapper/lubuntu-swap_1', '/dev/mapper/lubuntu-swap_1'))

        for test in tests:
            self.assertEqual(parse_swapoff(test[0]), test[1])

    @common.skipIfWindows
    def test_disable_swap_linux(self):
        """Test for disable_swap_linux() with mocks"""
        # No swap active - early return
        with mock.patch('bleachbit.Memory.count_swap_linux', return_value=0):
            self.assertIsNone(disable_swap_linux())

        # Command failure
        with mock.patch('bleachbit.Memory.count_swap_linux', return_value=1):
            with mock.patch('bleachbit.Memory.General.run_external', return_value=(1, '', 'swapoff failed')):
                self.assertRaisesRegex(
                    RuntimeError, 'swapoff failed', disable_swap_linux)

        # Unexpected output
        with mock.patch('bleachbit.Memory.count_swap_linux', return_value=1):
            with mock.patch('bleachbit.Memory.General.run_external', return_value=(0, 'unexpected\n', '')):
                self.assertRaisesRegex(
                    RuntimeError, 'Unexpected output', disable_swap_linux)

        # Parse success
        with mock.patch('bleachbit.Memory.count_swap_linux', return_value=1):
            with mock.patch('bleachbit.Memory.General.run_external', return_value=(0, 'swapoff /dev/sda3\n', '')):
                devices = disable_swap_linux()
                self.assertEqual(devices, ['/dev/sda3'])

    @common.skipIfWindows
    def test_enable_swap_linux(self):
        """Test for enable_swap_linux() with mocks"""
        # Success
        with mock.patch('bleachbit.Memory._', side_effect=lambda s: s):
            with mock.patch('bleachbit.Memory.General.run_external', return_value=(0, '', '')) as mock_run:
                enable_swap_linux()
                self.assertEqual(mock_run.call_args.args[0], ['swapon', '-a'])

        # Failure
        with mock.patch('bleachbit.Memory._', side_effect=lambda s: s):
            with mock.patch('bleachbit.Memory.General.run_external', return_value=(1, '', 'swapon failed')):
                self.assertRaisesRegex(
                    RuntimeError, 'swapon failed', enable_swap_linux)

    @common.skipIfWindows
    def test_get_swap_size_linux_errors(self):
        """Test for get_swap_size_linux() error paths"""
        # Malformed header
        self.assertRaisesRegex(
            RuntimeError, 'Unexpected first line',
            get_swap_size_linux, '/dev/sda1', 'bad header\n')

        # Missing device
        proc_swaps = 'Filename\tType\tSize\tUsed\tPriority\n/dev/sda2\tpartition\t123\t0\t-2\n'
        self.assertRaisesRegex(
            RuntimeError, 'cannot find size of swap device',
            get_swap_size_linux, '/dev/sda1', proc_swaps)

    @common.skipIfWindows
    def test_wipe_swap_linux(self):
        """Test for wipe_swap_linux() with mocks"""
        # devices is None - early return
        wipe_swap_linux(None, '')

        # Swap still in use
        with mock.patch('bleachbit.Memory.count_swap_linux', return_value=1):
            self.assertRaisesRegex(
                RuntimeError, 'Cannot wipe swap while it is in use',
                wipe_swap_linux, ['/dev/sda1'], '')

        # Safety limit exceeded
        safety_limit = 29 * 1024 ** 3
        with mock.patch('bleachbit.Memory.count_swap_linux', return_value=0):
            with mock.patch('bleachbit.Memory.get_swap_size_linux', return_value=safety_limit + 1):
                self.assertRaisesRegex(
                    RuntimeError, 'larger .* than expected',
                    wipe_swap_linux, ['/dev/sda1'], '')

        # UUID preservation
        with mock.patch('bleachbit.Memory.count_swap_linux', return_value=0):
            with mock.patch('bleachbit.Memory.get_swap_size_linux', return_value=1024 ** 2):
                with mock.patch('bleachbit.Memory.get_swap_uuid', return_value='abc-123'):
                    with mock.patch('bleachbit.Memory.wipe_contents'):
                        with mock.patch('bleachbit.Memory.General.run_external', return_value=(0, '', '')) as mock_run:
                            wipe_swap_linux(['/dev/sda1'], '')
                            args = mock_run.call_args[0][0]
                            self.assertIn('-U', args)
                            self.assertIn('abc-123', args)

        # mkswap failure
        with mock.patch('bleachbit.Memory.count_swap_linux', return_value=0):
            with mock.patch('bleachbit.Memory.get_swap_size_linux', return_value=1024 ** 2):
                with mock.patch('bleachbit.Memory.get_swap_uuid', return_value=None):
                    with mock.patch('bleachbit.Memory.wipe_contents'):
                        with mock.patch('bleachbit.Memory.General.run_external', return_value=(1, '', 'mkswap failed')):
                            self.assertRaisesRegex(
                                RuntimeError, 'mkswap failed',
                                wipe_swap_linux, ['/dev/sda1'], '')

    @common.skipIfWindows
    def test_wipe_memory(self):
        """Test for wipe_memory() with mocks"""
        # Command missing
        with mock.patch('bleachbit.FileUtilities.exe_exists', return_value=False):
            gen = wipe_memory()
            self.assertRaisesRegex(
                RuntimeError, 'Command swapon not found', next, gen)

        # Happy path
        with mock.patch('bleachbit.Memory._', side_effect=lambda s: s):
            with mock.patch('bleachbit.FileUtilities.exe_exists', return_value=True):
                with mock.patch('bleachbit.Memory.get_proc_swaps', return_value=''):
                    with mock.patch('bleachbit.Memory.disable_swap_linux', return_value=['/dev/sda1']):
                        with mock.patch('bleachbit.Memory.wipe_swap_linux'):
                            with mock.patch('os.fork', return_value=1):
                                with mock.patch('os.waitpid', return_value=(1, 0)):
                                    with mock.patch('bleachbit.Memory.enable_swap_linux'):
                                        gen = wipe_memory()
                                        self.assertTrue(next(gen))
                                        self.assertTrue(next(gen))
                                        self.assertEqual(next(gen), 0)

    @common.skipIfWindows
    def test_swap_off_swap_on(self):
        """Test for disabling and enabling swap"""
        if not common.have_root():
            self.skipTest('not enough privileges')
        disable_swap_linux()
        enable_swap_linux()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test case for module Cookie
"""

import hashlib
import json
import os
import sqlite3
import unittest
import shutil
from unittest import mock

import bleachbit
from bleachbit import Cookie
from bleachbit.Action import Cookie as CookieAction
from bleachbit.Cookie import COOKIE_KEEP_LIST_FILENAME, load_keep_list
from bleachbit.Options import options
from bleachbit.FileUtilities import execute_sqlite3
from tests import common


DEFAULT_COOKIE_FIXTURES = (
    ('example.com', 'session_id'),
    ('google.com', 'pref'),
    ('github.com', 'auth'),
)

DOMAIN_MATCHING_CASES = (
    ('example.com', 'example.com', True, 'Exact host match'),
    ('example.com', '.example.com', True,
     'Domain cookie canonicalizes to base host'),
    ('example.com', 'shop.example.com', True,
     'Subdomain should be preserved when base is keeplisted'),
    ('example.com', '.shop.example.com', True,
     'Firefox/Chromium store domain cookies with dots'),
    ('example.com', 'thisisanexample.com', False,
     'Substring matches must not count'),
    ('example.com', 'example.com.attacker.net', False,
     'Extra label to the right is a different domain'),
    ('example.com', 'shop-example.com', False,
     'Hyphenated host is unrelated'),
    ('www.example.com', 'example.com', False,
     'Host-only cookies should not match parent domain'),
    ('.example.com', 'example.com', True,
     'Leading dots in keep list should be ignored'),
    ('.example.com', 'www.example.com', True,
     'Leading dots in keep list should match subdomains'),
    ('.example.com', 'fooexample.com', False,
     'Canonical registrable domain mismatch'),
)


# Additional coverage:
# - Non-SQLite file with non-empty keep list should not be deleted (raises ValueError)
# - SQLite DB with unknown table and non-empty keep list should not be deleted (raises ValueError)
# - When no hosts match keep list, all rows are deleted and the database file is removed

class CookieTestCase(common.BleachbitTestCase):
    """Test cases for Cookie module"""

    def setUp(self):
        """Set up test fixtures"""
        super().setUp()
        # Reminder: each test case has its own, temporary bleachbit.ini config,
        # so no need to save/restore options.
        options.set('shred', True)

    def _get_file_hash(self, filepath):
        """Calculate SHA-256 hash of a file"""
        with open(filepath, 'rb') as f:
            return hashlib.sha256(f.read()).digest()

    def _sample_cookies(self, dotted=False):
        """Return a reusable copy of the default cookie fixtures"""
        cookies = DEFAULT_COOKIE_FIXTURES
        if dotted:
            return [
                (host if host.startswith('.') else f'.{host}', name)
                for host, name in cookies
            ]
        return [(host, name) for host, name in cookies]

    def _create_chrome_cookies_db(self, cookies=None):
        """Create a Chrome cookies database for testing

        Args:
            cookies: Iterable of (host_key, name) tuples. If None, uses
                default test cookies.

        Returns:
            str: Path to the created cookies database
        """
        if cookies is None:
            cookies = self._sample_cookies()

        path = os.path.join(self.tempdir, 'chrome_cookies.db')
        if os.path.exists(path):
            os.unlink(path)
        sql = '''
            CREATE TABLE cookies (
                host_key TEXT NOT NULL,
                name TEXT NOT NULL
            )
        '''
        execute_sqlite3(path, sql)

        conn = sqlite3.connect(path)
        try:
            cursor = conn.cursor()
            for host_key, name in cookies:
                cursor.execute(
                    'INSERT INTO cookies (host_key, name) VALUES (?, ?)',
                    (host_key, name),
                )
            conn.commit()
        finally:
            conn.close()

        return path

    def _find_live_firefox_cookies_file(self):
        """Return path to a real Firefox cookies.sqlite if one exists"""
        search_roots = [
            os.path.expanduser('~/snap/firefox/common/.mozilla/firefox'),
            os.path.expanduser('~/.mozilla/firefox'),
        ]
        appdata = os.environ.get('APPDATA')
        if appdata:
            search_roots.append(os.path.join(
                appdata, 'Mozilla', 'Firefox', 'Profiles'))

        for root in search_roots:
            if not root or not os.path.isdir(root):
                continue
            for dirpath, _dirnames, filenames in os.walk(root):
                if 'cookies.sqlite' in filenames:
                    return os.path.join(dirpath, 'cookies.sqlite')
        return None

    def _create_firefox_cookies_db(self, cookies=None):
        """Create a Firefox cookies database for testing

        Args:
            cookies: Iterable of cookie descriptors. Accepts:
                     * plain host strings
                     * (host, name) tuples
                     If None, uses default test cookies.

        Returns:
            str: Path to the created cookies database
        """
        if cookies is None:
            cookies = self._sample_cookies(dotted=True)

        path = os.path.join(self.tempdir, 'firefox_cookies.db')
        if os.path.exists(path):
            os.unlink(path)
        sql = '''
            CREATE TABLE moz_cookies (
                id INTEGER PRIMARY KEY,
                host TEXT NOT NULL,
                name TEXT NOT NULL,
                CONSTRAINT moz_uniqueid UNIQUE (host, name)
            )
        '''
        execute_sqlite3(path, sql)

        conn = sqlite3.connect(path)
        try:
            cursor = conn.cursor()
            for cookie in cookies:
                host = name = None
                if isinstance(cookie, str):
                    host, name = cookie, ''
                elif isinstance(cookie, (list, tuple)):
                    if len(cookie) >= 2:
                        host, name = cookie[:2]
                    elif len(cookie) == 1:
                        host, name = cookie[0], ''
                if host is None:
                    raise ValueError(
                        'Unsupported cookie format for Firefox test fixture')
                cursor.execute(
                    'INSERT INTO moz_cookies (host, name) VALUES (?, ?)', (host, name))
            conn.commit()
        finally:
            conn.close()

        return path

    def test_delete_cookie_value_error(self):
        """Empty keep list should raise ValueError (API requires non-empty set)"""
        path = self._create_chrome_cookies_db()

        # Get file hash before operation
        original_hash = self._get_file_hash(path)

        # Test scenarios that should raise ValueError
        scenarios = [
            # Scenario 1: Empty allowlist
            (set(), "Empty keep list should raise ValueError"),
            # Scenario 2: None allowlist
            (None, "None keep list should raise ValueError"),
        ]

        for keep_list, description in scenarios:
            with self.subTest(description=description):
                with self.assertRaises(ValueError):
                    Cookie.delete_cookies(path, keep_list, really_delete=True)
                self.assertExists(path)
                new_hash = self._get_file_hash(path)
                self.assertEqual(original_hash, new_hash,
                                 'Database file was modified')

    def test_delete_cookie_bad_file(self):
        """Empty keep list should raise ValueError (API requires non-empty set)"""
        path_missing = os.path.join(self.tempdir, 'nonexistent.db')
        path_non_sqlite = self.write_file(
            'experiment_626.dna', text='not sqlite3')
        path_wrong_sqlite = os.path.join(self.tempdir, 'wrong_sqlite.db')

        execute_sqlite3(
            path_wrong_sqlite, "CREATE TABLE not_browser_related (id INTEGER PRIMARY KEY)")

        scenarios = [
            (path_missing, "Non-existent file", ValueError),
            (path_non_sqlite, "Non-SQLite file", sqlite3.DatabaseError),
            (path_wrong_sqlite, "SQLite file with wrong table", ValueError),
        ]

        for test_path, description, expected_exception in scenarios:
            with self.subTest(description=description):
                original_hash = None
                if not test_path == path_missing:
                    original_hash = self._get_file_hash(test_path)
                with self.assertRaises(expected_exception):
                    Cookie.delete_cookies(
                        test_path, {'example.com'}, really_delete=True)

                if test_path == path_missing:
                    self.assertNotExists(test_path)
                else:
                    self.assertExists(test_path)
                    new_hash = self._get_file_hash(test_path)
                    self.assertEqual(original_hash, new_hash,
                                     'Database file was modified')

    def test_delete_cookies_chrome_with_keeplist(self):
        """Test selectively deleting Chrome cookies with keep list"""
        path = self._create_chrome_cookies_db()

        # Test deletion with keep list (keep google.com)
        keep_list = ['google.com']
        result = Cookie.delete_cookies(
            path, set(keep_list), really_delete=True)

        # example.com and github.com deleted
        self.assertEqual(result['total_deleted'], 2)
        self.assertEqual(result['total_kept'], 1)    # google.com kept
        self.assertFalse(result['skipped'])
        self.assertFalse(result['whole_file_deleted'])
        self.assertGreaterEqual(result['file_size_reduction'], 0)

    def test_non_sqlite_file_with_keeplist_raises(self):
        """Non-SQLite file should raise ValueError and not be deleted"""
        # First create a valid cookies db to get the path format right
        path = self._create_chrome_cookies_db()
        # Now remove it and create a non-sqlite file
        os.unlink(path)
        path = os.path.join(self.tempdir, 'not_sqlite.txt')
        with open(path, 'w', encoding='utf-8') as f:
            f.write('not a database')

        with self.assertRaises(sqlite3.DatabaseError):
            Cookie.delete_cookies(path, {'example.com'}, really_delete=True)

        self.assertTrue(os.path.exists(path))

    def test_sqlite_unknown_table_with_keeplist_raises(self):
        """SQLite DB without known cookies table should raise ValueError"""
        path = os.path.join(self.tempdir, 'unknown_table.db')
        conn = sqlite3.connect(path)
        cur = conn.cursor()
        cur.execute(
            'CREATE TABLE something_else (id INTEGER PRIMARY KEY, host TEXT)')
        conn.commit()
        conn.close()

        with self.assertRaises(ValueError):
            Cookie.delete_cookies(path, {'example.com'}, really_delete=True)

        self.assertTrue(os.path.exists(path))

    def test_delete_cookies_when_no_hosts_match_kept(self):
        """If no hosts match keep list, all rows are deleted and file is removed"""
        path = self._create_chrome_cookies_db()

        result = Cookie.delete_cookies(
            path, {'no-such-host.example'}, really_delete=True)

        self.assertEqual(result['total_kept'], 0)
        self.assertEqual(result['total_deleted'], 3)
        self.assertFalse(result['skipped'])
        self.assertTrue(result['whole_file_deleted'])
        self.assertGreaterEqual(result['file_size_reduction'], 0)

        # Verify database file was removed
        self.assertFalse(os.path.exists(path))

    def test_delete_cookies_firefox_with_keeplist(self):
        """Test selectively deleting Firefox cookies with keeplist"""
        path = self._create_firefox_cookies_db()

        # Test deletion with keep list
        keep_list = ['google.com']
        result = Cookie.delete_cookies(
            path, set(keep_list), really_delete=True)

        # example.com and github.com deleted
        self.assertEqual(result['total_deleted'], 2)
        self.assertEqual(result['total_kept'], 1)    # google.com kept
        self.assertFalse(result['skipped'])
        self.assertFalse(result['whole_file_deleted'])
        self.assertGreaterEqual(result['file_size_reduction'], 0)

        # Verify the database still exists and has the right cookies
        self.assertTrue(os.path.exists(path))

        conn = sqlite3.connect(path)
        cursor = conn.cursor()
        remaining_cookies = cursor.execute(
            'SELECT host FROM moz_cookies').fetchall()
        conn.close()

        self.assertEqual(len(remaining_cookies), 1)
        # Firefox stores domain cookies with a leading dot (e.g., '.google.com').
        self.assertEqual(remaining_cookies[0][0].lstrip('.'), 'google.com')

    def test_delete_firefox_live(self):
        """Test deleting Firefox cookies from a real database"""
        live_path = self._find_live_firefox_cookies_file()
        if not live_path:
            self.skipTest(
                'Skipping live Firefox test because cookies.sqlite was not found')

        def run_test_scenario(random_count):
            """Helper to test a cookie deletion scenario"""
            self.assertGreaterEqual(random_count, 0)
            temp_db = os.path.join(
                self.tempdir, f'firefox_live_copy{random_count}.sqlite')
            shutil.copy2(live_path, temp_db)

            # Copy WAL/SHM files if present to keep the database consistent
            for suffix in ('-wal', '-shm'):
                aux_src = f"{live_path}{suffix}"
                if os.path.exists(aux_src):
                    shutil.copy2(aux_src, f"{temp_db}{suffix}")

            # Add a non-existent domain to keep list to test filtering
            doesnotexist = 'doesnotexist1234567890.com'
            if random_count:
                conn = sqlite3.connect(temp_db)
                cursor = conn.cursor()
                random_hosts = cursor.execute(
                    f'SELECT DISTINCT host FROM moz_cookies ORDER BY RANDOM() LIMIT {random_count}').fetchall()
                conn.close()
                if not random_hosts:
                    self.skipTest('No hosts found in live Firefox database')
                keep_list = {host[0].lstrip('.') for host in random_hosts}
            else:
                keep_list = {doesnotexist}

            preview_result = Cookie.delete_cookies(
                temp_db, keep_list, really_delete=False)
            self.assertIn('total_deleted', preview_result)
            self.assertFalse(preview_result['skipped'])
            self.assertTrue(os.path.exists(temp_db))

            delete_result = Cookie.delete_cookies(
                temp_db, keep_list, really_delete=True)
            self.assertIn('total_deleted', delete_result)
            self.assertFalse(delete_result['skipped'])

            # Check file size reduction accuracy
            size_diff = abs(
                preview_result['file_size_reduction'] - delete_result['file_size_reduction'])
            # The estimate is accurate with shredding enabled, which is
            # how this test is run.
            self.assertLessEqual(
                size_diff, 1024, f"Size difference {size_diff:,} exceeds tolerance")

            if delete_result.get('whole_file_deleted'):
                self.assertFalse(os.path.exists(temp_db))
            else:
                self.assertTrue(os.path.exists(temp_db))
                os.unlink(temp_db)

        # Test scenario 1: Delete all cookies (keep non-existent domain)
        run_test_scenario(0)
        # Test scenario 2: Delete several random cookies
        run_test_scenario(10)
        run_test_scenario(50)

    def test_preview_cookies_deletion_chrome(self):
        """Test previewing Chrome cookie deletion"""
        path = self._create_chrome_cookies_db()

        # Test preview with keep list
        keep_list = ['google.com']
        result = Cookie.delete_cookies(
            path, set(keep_list), really_delete=False)

        self.assertEqual(result['total_deleted'], 2)
        self.assertEqual(result['total_kept'], 1)
        self.assertFalse(result['skipped'])
        self.assertFalse(result['whole_file_deleted'])
        self.assertGreaterEqual(result['file_size_reduction'], 0)

        # Verify original database is unchanged
        conn = sqlite3.connect(path)
        cursor = conn.cursor()
        all_cookies = cursor.execute('SELECT host_key FROM cookies').fetchall()
        conn.close()

        self.assertEqual(len(all_cookies), 3)  # All cookies still there

    def test_keeplist_domain_matching_matrix(self):
        """Table-driven coverage for keep list domain semantics"""
        for selection, host_value, expect_keep, reason in DOMAIN_MATCHING_CASES:
            with self.subTest(selection=selection, host=host_value, reason=reason):
                cookies = [(host_value, 'token')]
                path = self._create_chrome_cookies_db(cookies)
                result = Cookie.delete_cookies(
                    path, {selection}, really_delete=False)

                if expect_keep:
                    self.assertEqual(result['total_kept'], 1, reason)
                    self.assertEqual(result['total_deleted'], 0)
                else:
                    self.assertEqual(result['total_kept'], 0, reason)
                    self.assertEqual(result['total_deleted'], 1)

    def test_preview_cookies_deletion_no_keeplist(self):
        """Test previewing cookie deletion with no keep list"""
        # Create test cookies
        cookies_data = [
            ('example.com', 'session_id'),
        ]

        path = self._create_chrome_cookies_db(cookies_data)

        # Test preview with no keep list (should preview whole file deletion)
        with self.assertRaises(ValueError):
            Cookie.delete_cookies(path, set(), really_delete=False)

    def test_nonexistent_file(self):
        """Test handling of nonexistent cookie files"""
        path = os.path.join(self.tempdir, 'nonexistent.db')

        # Both deletion and preview should raise ValueError for nonexistent files
        with self.assertRaises(ValueError):
            Cookie.delete_cookies(path, {'example.com'}, really_delete=True)
        with self.assertRaises(ValueError):
            Cookie.delete_cookies(path, {'example.com'}, really_delete=False)

    def test_browser_detection(self):
        """Test automatic browser detection"""
        # Test Chrome detection
        chrome_path = self._create_chrome_cookies_db()
        self.assertEqual(Cookie.detect_browser(
            chrome_path), ('cookies', 'host_key'))

        # Test Firefox detection
        firefox_path = self._create_firefox_cookies_db()
        self.assertEqual(Cookie.detect_browser(
            firefox_path), ('moz_cookies', 'host'))

    def test_empty_database(self):
        """Test handling of empty cookie databases"""
        # Create empty Chrome database
        path = self._create_chrome_cookies_db([])

        # Test deletion with keep list
        result = Cookie.delete_cookies(
            path, {'example.com'}, really_delete=True)

        self.assertEqual(result['total_deleted'], 0)
        self.assertEqual(result['total_kept'], 0)
        self.assertFalse(result['skipped'])
        self.assertFalse(result['whole_file_deleted'])
        self.assertGreaterEqual(result['file_size_reduction'], 0)

    def _make_cookie_action(self):
        """Create a Cookie action instance with a mock XML element."""
        elem = mock.MagicMock()
        elem.getAttribute.return_value = ''
        elem.getAttribute.side_effect = lambda name: {
            'command': 'cookie',
            'search': 'file',
            'path': os.path.join(self.tempdir, 'nonexistent'),
        }.get(name, '')
        return CookieAction(elem)

    def test_load_keep_list_file_not_found(self):
        """FileNotFoundError should return empty set (no keep list configured)"""
        # Ensure the file doesn't exist
        keep_path = os.path.join(
            bleachbit.options_dir, COOKIE_KEEP_LIST_FILENAME)
        if os.path.exists(keep_path):
            os.remove(keep_path)
        result = load_keep_list()
        self.assertEqual(result, set())

    def test_load_keep_list_valid_json(self):
        """Valid JSON keep list should return the expected domains"""
        keep_data = ["example.com", "example.org"]
        keep_path = os.path.join(
            bleachbit.options_dir, COOKIE_KEEP_LIST_FILENAME)
        os.makedirs(bleachbit.options_dir, exist_ok=True)
        with open(keep_path, 'w', encoding='utf-8') as f:
            json.dump(keep_data, f)
        result = load_keep_list()
        self.assertEqual(result, {'example.com', 'example.org'})

    def test_load_keep_list_json_decode_error(self):
        """Corrupted JSON should raise JSONDecodeError, not return empty set"""
        action = self._make_cookie_action()
        keep_path = os.path.join(
            bleachbit.options_dir, COOKIE_KEEP_LIST_FILENAME)
        os.makedirs(bleachbit.options_dir, exist_ok=True)
        self.write_file(keep_path, '{invalid json', mode='w')
        with self.assertRaises(json.JSONDecodeError):
            load_keep_list()
        with self.assertRaises(json.JSONDecodeError):
            list(action.get_commands())

    def test_load_keep_list_unicode_decode_error(self):
        """UnicodeDecodeError should propagate, not return empty set"""
        action = self._make_cookie_action()
        keep_path = os.path.join(
            bleachbit.options_dir, COOKIE_KEEP_LIST_FILENAME)
        os.makedirs(bleachbit.options_dir, exist_ok=True)
        # Write invalid UTF-8 bytes
        with open(keep_path, 'wb') as f:
            f.write(b'\x80\x81\x82')
        with self.assertRaises(UnicodeDecodeError):
            load_keep_list()
        with self.assertRaises(UnicodeDecodeError):
            list(action.get_commands())

    def test_load_keep_list_permission_error(self):
        """PermissionError (a subtype of OSError) should propagate"""
        action = self._make_cookie_action()
        keep_path = os.path.join(
            bleachbit.options_dir, COOKIE_KEEP_LIST_FILENAME)
        os.makedirs(bleachbit.options_dir, exist_ok=True)
        self.write_file(keep_path, json.dumps(["example.com"]), mode='w')
        real_open = open

        def _mock_open(*args, **kwargs):
            if len(args) >= 1 and args[0] == keep_path:
                raise PermissionError('mock permission denied')
            return real_open(*args, **kwargs)
        with mock.patch('builtins.open', _mock_open):
            with self.assertRaises(PermissionError):
                load_keep_list()
            with self.assertRaises(PermissionError):
                list(action.get_commands())

    @common.skipIfWindows
    def test_load_keep_list_chmod(self):
        """Test that chmod affects file access"""
        action = self._make_cookie_action()
        keep_path = os.path.join(
            bleachbit.options_dir, COOKIE_KEEP_LIST_FILENAME)
        os.makedirs(bleachbit.options_dir, exist_ok=True)
        self.write_file(keep_path, json.dumps(["example.com"]), mode='w')
        # Remove read permission
        os.chmod(keep_path, 0o200)  # write-only
        with self.assertRaises(PermissionError):
            load_keep_list()
        with self.assertRaises(PermissionError):
            list(action.get_commands())
        os.chmod(keep_path, 0o600)  # restore read permission


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test case for module GuiStartup
"""

import os
import shutil
import stat
import subprocess
from unittest import mock

from bleachbit import IS_WINDOWS
import bleachbit
from bleachbit import GuiStartup
from bleachbit.Cleaner import register_cleaners
from bleachbit.GuiStartup import _has_selected_warning_option, _get_posix_permission_issues, _get_windows_permission_issues
from bleachbit.Options import options
from tests import common


class StubOptions:
    """Minimal stand-in for bleachbit.Options for testing."""

    def __init__(self, *, first_start=True, shred=False, old_version=None):
        """Seed stub with values."""
        self.values = {'first_start': first_start, 'shred': shred}
        self._old_version = old_version

    def get(self, key):
        """Return stored option value."""
        return self.values.get(key)

    def set(self, key, value):
        """Set a option value."""
        self.values[key] = value

    def get_tree(self, parent, child):
        """Return whether a tree option is selected."""
        return False


class GuiStartupTestCase(common.BleachbitTestCase):
    """Tests for GuiStartup helper logic."""

    def test_first_start_message_clears_flag(self):
        """Ensure first-start hint is emitted and flag gets cleared."""
        stub_options = StubOptions()

        with mock.patch.object(GuiStartup, 'options', stub_options), \
                mock.patch.object(GuiStartup, 'unset_sslkeylogfile', return_value=False):
            messages = GuiStartup.get_startup_messages(auto_exit=False)

        self.assertTrue(
            any('Access the application menu' in msg for msg, _ in messages))
        self.assertFalse(stub_options.get('first_start'))

    def test_has_selected_warning_option(self):
        """Test the function _has_selected_warning_option()"""
        # Register cleaners to populate backends
        list(register_cleaners())
        # Each test case starts with fresh options, so with
        # nothing selected, it should return False.
        self.assertFalse(_has_selected_warning_option())
        # Enable a safe option.
        options.set_tree('system', 'tmp', True)
        self.assertFalse(_has_selected_warning_option())
        # Enable an option with a warning.
        options.set_tree('system', 'empty_space', True)
        self.assertTrue(_has_selected_warning_option())
        # Clean up
        options.set_tree('system', 'tmp', False)
        options.set_tree('system', 'empty_space', False)

    def test_missing_requests(self):
        """Missing requests should be reported but not crash GuiStartup."""
        with common.mock_missing_package(
                'requests',
                clear_prefixes=('bleachbit.Network', 'bleachbit.Update', 'bleachbit.GuiStartup')):
            import bleachbit.GuiStartup as GuiStartup
            missing = GuiStartup._get_missing_dependencies()
            self.assertIn('requests', missing)

    @common.skipIfWindows
    def test_is_config_writable(self):
        """Test read-only configuration file"""
        o = bleachbit.Options.Options()
        o.close()
        self.assertExists(bleachbit.options_file)

        if common.have_root():
            # root bypasses file permission checks, so use chattr +i
            subprocess.run(
                ['chattr', '+i', bleachbit.options_file], check=True)
        else:
            os.chmod(bleachbit.options_file, stat.S_IRUSR |
                     stat.S_IRGRP | stat.S_IROTH)
        try:
            issues = GuiStartup._get_config_permission_issues()
        finally:
            if common.have_root():
                subprocess.run(
                    ['chattr', '-i', bleachbit.options_file], check=True)
            os.unlink(bleachbit.options_file)

        self.assertNotExists(bleachbit.options_file)
        self.assertIsInstance(issues, list)
        self.assertGreater(len(issues), 0)
        self.assertTrue(any('Write error' in issue for issue in issues),
                        f"Expected 'Write error' in issues: {issues}")

    def test_config_permission_issues_normal(self):
        """Test that a normal writable config file reports no issues."""
        o = bleachbit.Options.Options()
        o.close()
        self.assertExists(bleachbit.options_file)
        issues = GuiStartup._get_config_permission_issues()
        os.unlink(bleachbit.options_file)
        self.assertNotExists(bleachbit.options_file)
        self.assertFalse(issues)

    def test_config_permission_nonissue_missing_file(self):
        """A missing config file must not be reported as an error.

        Unlike another test, this scenario has an options directory
        that allows creating the missing config file.
        """
        if os.path.exists(bleachbit.options_file):
            os.unlink(bleachbit.options_file)
        self.assertNotExists(bleachbit.options_file)
        issues = GuiStartup._get_config_permission_issues()
        self.assertFalse(issues)

    @common.skipIfWindows
    def test_config_permission_issue_non_writeable_options_dir(self):
        """Config does not exist and options dir is not writeable"""
        non_writable_dir = self.mkdir('not_writeable_options')
        non_writable_file = os.path.join(non_writable_dir, 'bleachbit.ini')
        use_chattr = common.have_root()
        try:
            if use_chattr:
                # Root bypasses chmod, so use chattr +i instead
                subprocess.run(['chattr', '+i', non_writable_dir], check=True)
            else:
                os.chmod(non_writable_dir, stat.S_IRUSR | stat.S_IXUSR)
            self.assertNotExists(non_writable_file)
            with mock.patch('bleachbit.options_dir', non_writable_dir), \
                    mock.patch('bleachbit.options_file', non_writable_file):
                issues = GuiStartup._get_config_permission_issues()
        finally:
            if use_chattr:
                subprocess.run(['chattr', '-i', non_writable_dir], check=True)
            else:
                os.chmod(non_writable_dir, stat.S_IRWXU)
            shutil.rmtree(non_writable_dir)
        self.assertIsInstance(issues, list)
        self.assertGreater(len(issues), 0)
        self.assertTrue(any('Write error' in issue for issue in issues),
                        f"Expected 'Write error' in issues: {issues}")

    @common.also_with_sudo
    def test_permission_issues_normal_file(self):
        """Test ownership check on a normal file owned by the user."""
        path = self.write_file('check_me')
        fstat = os.stat(path)
        if IS_WINDOWS:
            has_error, lines = _get_windows_permission_issues(path)
        else:
            has_error, lines = _get_posix_permission_issues(fstat, path)
        self.assertFalse(has_error)
        self.assertTrue(any('File owner:' in line for line in lines))
        self.assertTrue(any('Current user:' in line for line in lines))

    @common.skipUnlessWindows
    def test_get_windows_user_info(self):
        """Test _get_windows_user_info function."""
        current_sid, current_name, group_sids = GuiStartup._get_windows_user_info()
        self.assertIsNotNone(current_sid)
        self.assertIsNotNone(current_name)
        self.assertIsNotNone(group_sids)
        self.assertIsInstance(current_sid, str)
        self.assertIsInstance(current_name, str)
        self.assertIsInstance(group_sids, set)
        self.assertEqual(current_name, os.environ.get('USERNAME', ''))

    @common.skipUnlessWindows
    def test_get_windows_file_owner(self):
        """Test _get_windows_file_owner function."""
        owner_sid_str, owner_name = GuiStartup._get_windows_file_owner(
            self.tempdir)
        self.assertIsNotNone(owner_sid_str)
        self.assertIsNotNone(owner_name)
        self.assertIsInstance(owner_sid_str, str)
        self.assertIsInstance(owner_name, str)
        if 'APPVEYOR' in os.environ:
            expected_owner = 'Administrators'
        else:
            expected_owner = os.environ.get('USERNAME', '')
        self.assertEqual(owner_name, expected_owner)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test cases for module Winapp
"""

import os
import shutil
import stat
import string
import struct
import tempfile
import time
from unittest import mock

from tests import common
from bleachbit.Winapp import Winapp, detectos, detect_file, section2option
from bleachbit.Windows import detect_registry_key, parse_windows_build
from bleachbit import IS_POSIX, IS_WINDOWS, logger

if IS_WINDOWS:
    import winreg

def _create_registry_keys(*key_paths):
    """Create registry keys, ignoring errors if they already exist"""
    if not IS_WINDOWS:
        return
    for key_path in key_paths:
        try:
            hkey = winreg.CreateKey(winreg.HKEY_CURRENT_USER, key_path)
            hkey.Close()
        except OSError:
            pass

def _delete_registry_keys(*key_paths):
    """Delete registry keys, ignoring errors if they don't exist"""
    if not IS_WINDOWS:
        return
    for key_path in key_paths:
        try:
            winreg.DeleteKey(winreg.HKEY_CURRENT_USER, key_path)
        except OSError:
            pass

KEYFULL = 'HKCU\\Software\\BleachBit\\DeleteThisKey'


def get_winapp2():
    """Download and cache winapp2.ini.  Return local filename."""
    url = ("https://raw.githubusercontent.com/bleachbit/winapp2.ini"
            "/refs/heads/master/Winapp2-BleachBit.ini")
    tmpdir = None
    if IS_POSIX:
        tmpdir = '/tmp'
    if IS_WINDOWS:
        tmpdir = os.getenv('TMP')
    if not tmpdir:
        tmpdir = tempfile.gettempdir()
    fname = os.path.join(tmpdir, 'bleachbit_test_winapp2.ini')
    if os.path.exists(fname):
        age_seconds = time.time() - os.stat(fname)[stat.ST_MTIME]
        if age_seconds > (24 * 36 * 36):
            logger.info('deleting stale file %s ', fname)
            os.remove(fname)
    if not os.path.exists(fname):
        from bleachbit.Network import download_url_to_fn
        assert download_url_to_fn(url, fname)
    assert os.path.exists(fname)
    return fname


class WinappTestCase(common.BleachbitTestCase):
    """Test cases for Winapp"""

    ini_fn = None

    def run_all(self, cleaner, really_delete):
        """Test all the cleaner options"""
        for (option_id, __name) in cleaner.get_options():
            for cmd in cleaner.get_commands(option_id):
                for result in cmd.execute(really_delete):
                    common.validate_result(self, result, really_delete)

    @common.skipUnlessWindows
    def test_remote(self):
        """Test with downloaded file"""
        winapps = Winapp(get_winapp2())
        for cleaner in winapps.get_cleaners():
            self.run_all(cleaner, False)

    def test_detectos(self):
        """Test detectos function"""
        # Tests are in the format (required_ver, mock, expected_return)
        tests = (('5.1', '5.1', True),
                 ('5.1', '6.0', False),
                 ('6.0', '5.1', False),
                 # 5.1 is the maximum version
                 ('|5.1', '5.1', True),
                 ('|5.1', '6.0', False),
                 ('|5.1', '10.0', False),
                 # 10.0 is the maximum version
                 ('|10.0', '5.1', True),
                 ('|10.0', '10.0', True),
                 ('|10.0', '10.1', False),
                 # 6.1 is the minimum version
                 ('6.1|', '5.1', False),
                 ('6.1|', '6.0', False),
                 ('6.1|', '6.1', True),
                 ('6.1|', '6.2', True),
                 ('6.1|', '10.0', True),
                 ('6.2|', '5.1', False),
                 ('6.2|', '6.0', False),
                 ('6.2|', '6.1', False),
                 ('6.2|', '6.2', True),
                 # must be 6.2 or 6.3
                 ('6.2|6.3', '6.0', False),
                 ('6.2|6.3', '6.1', False),
                 ('6.2|6.3', '6.2', True),
                 ('6.2|6.3', '6.3', True),
                 ('6.2|6.3', '10.0', False),
                 # 10.0 is the minimum
                 ('10.0|', '5.1', False),
                 ('10.0|', '10.0', True))
        for (req, mock_ver, expected_return) in tests:
            mock_ver = parse_windows_build(mock_ver)
            actual_return = detectos(req, mock_ver)
            msg = (f'detectos({req}, {mock_ver})=={actual_return}'
                   f' instead of {expected_return}')
            self.assertEqual(expected_return, actual_return, msg)

    @common.skipUnlessWindows
    def test_detect_file(self):
        """Test detect_file function"""
        tests = [('%windir%\\system32\\kernel32.dll', True),
                 ('%windir%\\system32', True),
                 ('%ProgramFiles%\\Internet Explorer', True),
                 ('%ProgramFiles%\\Internet Explorer\\', True),
                 ('%windir%\\doesnotexist', False),
                 ('%windir%\\system*', True),
                 ('%windir%\\*ystem32', True),
                 ('%windir%\\*ystem3*', True)]
        # On 64-bit Windows, Winapp2.ini expands the %ProgramFiles% environment
        # variable to also %ProgramW6432%, so test unique entries in
        # %ProgramW6432%.
        if os.getenv('ProgramW6432'):
            dir_64 = os.listdir(os.getenv('ProgramFiles'))
            dir_32 = os.listdir(os.getenv('ProgramW6432'))
            dir_32_unique = set(dir_32) - set(dir_64)
            if dir_32 and not dir_32_unique and 8 * struct.calcsize('P') == 32:
                raise RuntimeError(
                    'Test expects objects in %ProgramW6432% not in %ProgramFiles%')
            for pathname in dir_32_unique:
                tests.append((f'%ProgramFiles%\\{pathname}', True))
        else:
            logger.info(
                'skipping %ProgramW6432% tests because WoW64 not detected')
        for pathname, expected_return in tests:
            actual_return = detect_file(pathname)
            msg = f'detect_file({pathname}) returned {actual_return}'
            self.assertEqual(expected_return, actual_return, msg)

    def setup_fake(self, f1_filename=None):
        """Setup the test environment"""
        subkey = 'Software\\BleachBit\\DeleteThisKey\\AndThisKey'

        # put ampersand in directory name to test
        # https://github.com/bleachbit/bleachbit/issues/308
        dirname = self.mkdtemp(prefix='bleachbit-test-winapp&')
        fname1 = self.write_file(os.path.join(dirname, f1_filename or 'deleteme.log'), b'', 'wb')

        dirname2 = self.mkdir(os.path.join(dirname, 'sub'))
        fname2 = self.write_file(os.path.join(dirname2, 'deleteme.log'), b'', 'wb')

        fbak = self.write_file(os.path.join(dirname, 'deleteme.bak'), b'', 'wb')

        self.assertExists(fname1)
        self.assertExists(fname2)
        self.assertExists(fbak)

        _create_registry_keys(subkey)

        self.assertTrue(detect_registry_key(KEYFULL))
        self.assertTrue(detect_registry_key(f'HKCU\\{subkey}'))

        return dirname, fname1, fname2, fbak

    def ini2cleaner(self, body, do_next=True):
        """Write a minimal Winapp2.ini"""
        with open(self.ini_fn, 'w', encoding='utf-8') as ini:
            ini.write('[someapp]\nLangSecRef=3021\n' + body + '\n')
        self.assertExists(self.ini_fn)
        if do_next:
            return next(Winapp(self.ini_fn).get_cleaners())
        return Winapp(self.ini_fn).get_cleaners()

    @common.skipUnlessWindows
    def test_fake(self):
        """Test with fake file"""

        # reuse this path to store a winapp2.ini file in
        self.ini_fn = self.mkstemp(suffix='.ini', prefix='winapp2')

        # a set of tests
        # this map explains what each position in the test tuple means
        # 0=line to write directly to winapp2.ini
        # 1=filename1 to place in fake environment (default=deleteme.log)
        # 2=auto-hide before cleaning
        # 3=dirname exists after cleaning
        # 4=filename1 (.\deleteme.log) exists after cleaning
        # 5=sub\deleteme.log exists after cleaning
        # 6=.\deleteme.bak exists after cleaning
        tests = [
            # single file
            ('FileKey1=%s|deleteme.log', None,
                False, True, False, True, True),
            # single file, case matching should be insensitive
            ('FileKey1=%s|dEleteme.LOG', None,
                False, True, False, True, True),
            # special characters for XML
            ('FileKey1=%s|special_chars_&-\'.txt', 'special_chars_&-\'.txt',
             False, True, False, True, True),
            # *.log
            ('FileKey1=%s|*.LOG', None, False, True, False, True, True),
            # semicolon separates different file types
            ('FileKey1=%s|*.log;*.bak', None,
             False, True, False, True, False),
            # *.*
            ('FileKey1=%s|*.*', None, False, True, False, True, False),
            # recurse *.*
            ('FileKey1=%s|*.*|RECURSE', None, False,
             True, False, False, False),
            # recurse *.log
            ('FileKey1=%s|*.log|RECURSE', None, False,
             True, False, False, True),
            # remove self *.*, this removes the directory
            ('FileKey1=%s|*.*|REMOVESELF', None,
             False, False, False, False, False),
        ]

        # Add positive detection, where the detection believes the application is present,
        # to all the tests, which are also positive.
        new_tests = []
        for test in tests:
            for detect in (
                "\nDetectFile=%%APPDATA%%\\Microsoft",
                "\nSpecialDetect=DET_WINDOWS",
                "\nDetectFile1=%%APPDATA%%\\Microsoft\nDetectFile2=%%APPDATA%%\\does_not_exist",
                "\nDetectFile1=%%APPDATA%%\\does_not_exist\nDetectFile2=%%APPDATA%%\\Microsoft",
                "\nDetect=HKCU\\Software\\Microsoft",
                # Below checks that a space is OK in the registry key
                "\nDetect=HKCU\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon",
                # Below checks Detect# where one of two keys exist.
                "\nDetect1=HKCU\\Software\\Microsoft\nDetect2=HKCU\\Software\\does_not_exist",
                "\nDetect1=HKCU\\Software\\does_not_exist\nDetect2=HKCU\\Software\\Microsoft",
                # Below checks Detect with DetectFile where one exists
                "\nDetect=HKCU\\Software\\Microsoft\nDetectFile=%%APPDATA%%\\does_not_exist",
                    "\nDetect=HKCU\\Software\\does_not_exist\nDetectFile=%%APPDATA%%\\Microsoft"):
                new_ini = test[0] + detect
                new_test = [new_ini, ] + list(test[1:])
                new_tests.append(new_test)
        positive_tests = tests + new_tests

        # execute positive tests
        for test in positive_tests:
            print('positive test: ', test)
            self.assertEqual(len(test), 7)
            (dirname, fname1, fname2, fbak) = self.setup_fake(test[1])
            cleaner = self.ini2cleaner(test[0] % dirname)
            self.assertEqual(test[2], cleaner.auto_hide())
            self.run_all(cleaner, False)
            self.run_all(cleaner, True)
            self.assertCondExists(test[3], dirname)
            self.assertCondExists(test[4], fname1)
            self.assertCondExists(test[5], fname2)
            self.assertCondExists(test[6], fbak)
            shutil.rmtree(dirname, True)

        # negative tests where the application detect believes the application
        # is absent
        for test in tests:
            for detect in (
                "\nDetectFile=c:\\does_not_exist",
                "\nSpecialDetect=DET_SPACE_QUEST",
                # special characters for XML
                "\nDetectFile=c:\\does_not_exist_special_chars_&'",
                "\nDetectFile1=c:\\does_not_exist1\nDetectFile2=c:\\does_not_exist2",
                "\nDetect=HKCU\\Software\\does_not_exist",
                "\nDetect=HKCU\\Software\\does_not_exist_&'",
                    "\nDetect1=HKCU\\Software\\does_not_exist1"
                    "\nDetect2=HKCU\\Software\\does_not_exist1"):
                new_ini = test[0] + detect
                test_full = [new_ini, ] + list(test[1:])
                print('negative test', test_full)
                # execute the test
                (dirname, fname1, fname2, fbak) = self.setup_fake()
                cleaner = self.ini2cleaner(test_full[0] % dirname, False)
                self.assertRaises(StopIteration, cleaner.__next__)
                shutil.rmtree(dirname, True)

        # registry key, basic
        (dirname, fname1, fname2, fbak) = self.setup_fake()
        cleaner = self.ini2cleaner(f'RegKey1={KEYFULL}')
        self.run_all(cleaner, False)
        self.assertTrue(detect_registry_key(KEYFULL))
        self.run_all(cleaner, True)
        self.assertFalse(detect_registry_key(KEYFULL))
        shutil.rmtree(dirname, True)

        # check for parse error with ampersand
        (dirname, fname1, fname2, fbak) = self.setup_fake()
        cleaner = self.ini2cleaner(
            'RegKey1=HKCU\\Software\\PeanutButter&Jelly')
        self.run_all(cleaner, False)
        self.run_all(cleaner, True)
        shutil.rmtree(dirname, True)

    @common.skipUnlessWindows
    def test_excludekey(self):
        """Test for ExcludeKey"""

        # reuse this path to store a winapp2.ini file in
        self.ini_fn = self.mkstemp(suffix='.ini', prefix='winapp2')

        # Each tuple contains:
        #   [0] = Body of winapp2.ini configuration
        #   [1] = Whether .\deleteme.log should exist after cleaning (True=exists)
        #   [2] = Whether .\deleteme.bak should exist after cleaning (True=exists)
        #   [3] = Whether sub\deleteme.log should exist after cleaning (True=exists)
        #
        # Path notation:
        #   %(d)s     - Literal test directory path without a trailing backslash
        #   %(d)s\\   - Literal directory with a trailing backslash
        #   %%bbtestdir%% - Literal environment variable that resolves to same test directory

        tests = (
            # delete everything in single directory (no children) without
            # exclusions
            ('FileKey1=%(d)s|deleteme.*', False, False, True),
            # delete everything in single directory using environment variable
            ('FileKey1=%%bbtestdir%%|deleteme.*', False, False, True),
            # delete everything in parent and child directories without
            # exclusions
            ('FileKey1=%(d)s|deleteme.*|RECURSE', False, False, False),
            # don't delete files containing the described file name
            ('FileKey1=%(d)s|eteme.*|RECURSE', True, True, True),
            # delete a pattern sub folder included
            ('FileKey1=%(d)s|*eteme.*|RECURSE', False, False, False),
            # delete a pattern sub folder not included
            ('FileKey1=%(d)s|*eteme.*', False, False, True),
            # exclude log delimited by pipe
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=FILE|%(d)s|deleteme.log',
             True, False, True),
            # trailing backslash should work the same as without
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=FILE|%(d)s\\|deleteme.log',
             True, False, True),
            # delete / exclude patterns with different complexity
            ('FileKey1=%(d)s|*eteme.*\nExcludeKey1=FILE|%(d)s|deleteme.log',
             True, False, True),
            ('FileKey1=%(d)s|*eteme.*\nExcludeKey1=FILE|%(d)s\\|deleteme.log',
             True, False, True),
            ('FileKey1=%(d)s|*ete*.*\nExcludeKey1=PATH|%(d)s|*ete*.lo?',
             True, False, True),
            ('FileKey1=%(d)s|*eteme*\nExcludeKey1=PATH|%(d)s|*notexist.*',
             False, False, True),
            ('FileKey1=%(d)s|*eteme.log\nExcludeKey1=PATH|%(d)s|do*.log',
             False, True, True),
            # semicolon separates different file types
            ('FileKey1=%(d)s|*.log;*.bak|RECURSE\nExcludeKey1=PATH|%(d)s|*.log',
             True, False, True),
            # exclude log without pipe delimiter
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=FILE|%(d)s\\deleteme.log',
             True, False, True),
            # exclude everything in folder
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=PATH|%(d)s|*.*',
             True, True, True),
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=PATH|%(d)s\\|*.*',
             True, True, True),
            ('FileKey1=%(d)s|deleteme.*|RECURSE\nExcludeKey1=PATH|%(d)s|*.*',
             True, True, True),
            ('FileKey1=%(d)s|*.*|RECURSE\nExcludeKey1=PATH|%(d)s|*.*',
             True, True, True),
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=PATH|%(d)s',
             True, True, True),
            # exclude everything in folder using environment variable
            # use double %% to escape
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=PATH|%%bbtestdir%%|*.*',
             True, True, True),
            # exclude sub-folder
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=PATH|%(d)s\\sub',
             False, False, True),
            # exclude multiple file types that do not exist
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=PATH|%(d)s|*.exe;*.dll',
             False, False, True),
            # exclude multiple file types that do exist, so dlete nothing
            ('FileKey1=%(d)s|deleteme.*\nExcludeKey1=PATH|%(d)s|*.bak;*.log',
             True, True, True),
            ('FileKey1=%(d)s|deleteme.*|RECURSE\nExcludeKey1=PATH|%(d)s|*.bak;*.log',
             True, True, True),
            # multiple ExcludeKey, neither of which should do anything
            ('FileKey1=%(d)s|deleteme.*|RECURSE'
             '\nExcludeKey1=PATH|c:\\doesnotexist|*.*'
             '\nExcludeKey2=PATH|c:\\alsodoesnotexist\\|*.*',
             False, False, False),
            # multiple ExcludeKey, the first should work
            ('FileKey1=%(d)s|deleteme.*|RECURSE'
             '\nExcludeKey1=PATH|%(d)s|*.log'
             '\nExcludeKey2=PATH|c:\\alsodoesnotexist\\|*.*',
             True, False, True),
            # multiple ExcludeKey, both should work
            ('FileKey1=%(d)s|deleteme.*|RECURSE'
             '\nExcludeKey1=PATH|%(d)s|*.log'
             '\nExcludeKey2=PATH|%(d)s|*.bak',
             True, True, True),
            # glob should exclude the directory called 'sub'
            ('FileKey1=%(d)s|*.*\nExcludeKey1=PATH|%(d)s\\s*',
             False, False, True),
        )

        for test in tests:
            msg = f'\nTest:\n{test[0]}'
            # setup
            (dirname, _fname1, _fname2, _fbak) = self.setup_fake()
            self.assertExists(rf'{dirname}\deleteme.log', msg)
            self.assertExists(rf'{dirname}\deleteme.bak', msg)
            self.assertExists(rf'{dirname}\sub\deleteme.log', msg)
            # set environment variable for testing
            with common.set_temporary_env('bbtestdir', dirname):
                self.assertExists(r'$bbtestdir\deleteme.log', msg)
                # delete files
                cleaner = self.ini2cleaner(test[0] % {'d': dirname})
                self.run_all(cleaner, True)
                # test
                self.assertCondExists(test[1], rf'{dirname}\deleteme.log', msg)
                self.assertCondExists(test[2], rf'{dirname}\deleteme.bak', msg)
                self.assertCondExists(
                    test[3], rf'{dirname}\sub\deleteme.log', msg)

            # cleanup
            shutil.rmtree(dirname, True)

    def _verify_keys_state(self, expected_state):
        """Verify registry keys match expected state (dict of key_path -> exists)"""
        for key_path, should_exist in expected_state.items():
            exists = detect_registry_key('HKCU\\' + key_path)
            if should_exist:
                self.assertTrue(exists, f'Key {key_path} should exist')
            else:
                self.assertFalse(exists, f'Key {key_path} should not exist')

    @common.skipUnlessWindows
    def test_excludekey_reg(self):
        """Test for ExcludeKey with REG type to prevent registry key deletion"""

        # reuse this path to store a winapp2.ini file in
        self.ini_fn = self.mkstemp(suffix='.ini', prefix='winapp2')

        test_key_parent = r'Software\BleachBitTest\ExcludeKey'
        test_key_child0 = rf'{test_key_parent}\Child0'
        test_key_child1 = rf'{test_key_parent}\Child1'
        test_key_not_exc = f'{test_key_parent}_NotExcluded'

        tests = (
            # (exclude_key, parent_should_remain, child0_should_remain, child1_should_remain)
            # Excluding a child key also implicitly protects the parent,
            # since the parent must exist for the excluded child to persist.
            (test_key_parent, True, True, True),
            (test_key_child0, True, True, False),
            (test_key_child1, True, False, True),
        )

        for exclude_key, parent_should_remain, child0_should_remain, child1_should_remain in tests:
            with self.subTest(exclude_key=exclude_key):
                try:
                    _create_registry_keys(
                        test_key_parent, test_key_child0,
                        test_key_child1, test_key_not_exc)
                    cleaner_ini = f'''RegKey1=HKCU\\{test_key_parent}
RegKey2=HKCU\\{test_key_child0}
RegKey3=HKCU\\{test_key_child1}
RegKey4=HKCU\\{test_key_not_exc}
ExcludeKey1=REG|HKCU\\{exclude_key}'''
                    cleaner = self.ini2cleaner(cleaner_ini)
                    self.run_all(cleaner, False)
                    self._verify_keys_state({
                        test_key_parent: True,
                        test_key_child0: True,
                        test_key_child1: True,
                        test_key_not_exc: True,
                    })
                    self.run_all(cleaner, True)
                    self._verify_keys_state({
                        test_key_parent: parent_should_remain,
                        test_key_child0: child0_should_remain,
                        test_key_child1: child1_should_remain,
                        test_key_not_exc: False,
                    })
                finally:
                    _delete_registry_keys(
                        test_key_parent, test_key_child0,
                        test_key_child1, test_key_not_exc)

    @common.skipUnlessWindows
    def test_filekey_with_path_including_systemdrive(self):
        """Test FileKey with path including SystemDrive"""
        self.ini_fn = self.mkstemp(suffix='.ini', prefix='winapp2')
        # lexists(r'C:filename') returns False
        # FindFilesW(r'C:filename') throws an exception
        # So assert that we use r'C:\filename'.
        # We need this test because winapp2.ini supports such paths and
        # giving r'C:filename' as argument to lexists or FindFilesW
        # causes a hard to detect issue, which has been detected because
        # lexists('C:filename') returns True when called through PyCharm.
        # Also test_remote is supposed to run cleaners that contain
        # such paths but it doesn't assert if those cleaners actually
        # clean their targets.

        filename = os.path.join('C:\\', 'deleteme.txt')
        try:
            common.touch_file(filename)
        except PermissionError:
            self.skipTest('Permission denied: run as administrator')

        with mock.patch('os.path.lexists') as mock_lexists:
            cleaner = self.ini2cleaner('FileKey1=%SystemDrive%|deleteme.txt')
            self.run_all(cleaner, True)
            mock_lexists.assert_any_call(r'C:\deleteme.txt')

        self.assertNotExists(filename)

    @common.skipUnlessWindows
    def test_removeself(self):
        """Test for the removeself option"""

        # indexes for test
        # position 0: FileKey statement
        # position 1: whether the file `dir_c\subdir\foo.log` should exist
        #              after operation is complete
        # position 2: path of top-folder which should have been deleted
        d = self.tempdir
        tests = (
            # Refer to directory directly (i.e., without a glob).
            (rf'FileKey1={d}\dir_c|*.*|REMOVESELF',
             False, rf'{d}\dir_c'),
            # Refer to file that exists. This is invalid, so nothing happens.
            (rf'FileKey1={d}\dir_c\submarine_sandwich.log|*.*|REMOVESELF',
             True, ''),
            (rf'FileKey1={d}\dir_c\submarine*|*.*|REMOVESELF',
             True, ''),
            # Refer to path that does not exist, so nothing happens.
            (rf'FileKey1={d}\dir_c\doesnotexist.log|*.*|REMOVESELF',
             True, ''),
            # Refer by glob to both a file and directory (which both start with `sub`).
            # This should affect only the directory.
            (rf'FileKey1={d}\dir_a\sub*|*.*|REMOVESELF',
             True, rf'{d}\dir_a\subdir'),
            # glob in middle of directory path with whole directory entry
            (rf'FileKey1={d}\*c\subdir|*.*|REMOVESELF',
             False, rf'{d}\dir_c\subdir'),
            (rf'FileKey1={d}\*doesnotexist\subdir|*.*|REMOVESELF',
             True, ''),
            # glob at end of path
            (rf'FileKey1={d}\dir_c\sub*|*.*|REMOVESELF',
             False, rf'{d}\dir_c\subdir'),
            (rf'FileKey1={d}\dir_c\doesnotexist*|*.*|REMOVESELF',
             True, '')
        )

        self.ini_fn = self.mkstemp(suffix='.ini', prefix='winapp2')

        for filekey, c_log_expected, top_log_expected in tests:
            for letter in ('a', 'b', 'c'):
                # Make three directories, each with a `foo.log` file.
                fn = os.path.join(self.tempdir, 'dir_' +
                                  letter, 'subdir', 'foo.log')
                common.touch_file(fn)

            # In dir_a, place a file one level up from `foo.log`.
            # Notice the glob `dir_a\sub*` will match both a directory and file.
            fn2 = os.path.join(self.tempdir, 'dir_a', 'submarine_sandwich.log')
            common.touch_file(fn2)

            cleaner = self.ini2cleaner(filekey)
            self.assertExists(fn, filekey)
            self.assertExists(fn2, filekey)
            if top_log_expected:
                self.assertExists(top_log_expected, filekey)
            self.run_all(cleaner, True)
            if c_log_expected:
                self.assertExists(fn, filekey)
            else:
                self.assertNotExists(fn, filekey)
            if top_log_expected != '':
                self.assertNotExists(top_log_expected, filekey)
            self.assertExists(fn2, filekey)
            if top_log_expected:
                self.assertNotExists(top_log_expected, filekey)

    def test_section2option(self):
        """Test for section2option()"""
        tests = (('  FOO2  ', 'foo2'),
                 ('A - B (C)', 'a_b_c'))
        for test in tests:
            self.assertEqual(section2option(test[0]), test[1])

    def test_section_not_found(self):
        """Test a section that is found"""
        tmp_ini_fn = self.write_file(
            'section_not_found.ini',
            r"""[Initech TPS Reports *]
LangSecRef=3023
DetectFile=%CommonAppData%\Initech\TPSReports
FileKey1=%CommonAppData%\Initech\TPSReports\Logs|*.*""",
            mode='w')

        with mock.patch.object(Winapp, 'detect', return_value=True) as mock_detect:
            winapps = Winapp(tmp_ini_fn)
            cleaner_list = list(winapps.get_cleaners())
            self.assertEqual(len(cleaner_list), 1)
            self.assertEqual(cleaner_list[0].get_name(), 'Multimedia')
            this_option = list(cleaner_list[0].get_options())[0]
            self.assertEqual(this_option[0], 'initech_tps_reports')
            self.assertEqual(this_option[1], 'Initech TPS Reports')
            mock_detect.assert_called()

    def test_many_patterns(self):
        """Test a cleaner like Steam Installers and related performance improvement

        https://github.com/bleachbit/bleachbit/issues/325
        """

        # set up environment
        file_count = 1000
        dir_count = 10
        print(f'Making {file_count} files in each of {dir_count} directories.')
        for i_d in range(1, dir_count + 1):
            sub_dir = self.mkdir(os.path.join(self.tempdir, f'dir{i_d}'))
            for i_f in range(1, file_count + 1):
                self.write_file(os.path.join(sub_dir, f'file{i_f}'))

        self.ini_fn = self.mkstemp(suffix='.ini', prefix='winapp2')

        searches = ';'.join(
            [f'*.{letter}' for letter in string.ascii_letters[0:26]])
        cleaner = self.ini2cleaner(
            f'FileKey1={self.tempdir}|{searches}|RECURSE')

        # preview
        t0 = time.time()
        self.run_all(cleaner, False)
        t1 = time.time()
        print(f'Elapsed time in preview: {t1 - t0:.4f} seconds ')

        # delete
        self.run_all(cleaner, True)


# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test case for the module Process
"""
import os
import sys
from unittest import mock

# While POSIX systems may not always have psutil,
# the test environment (and Windows) always should have it.
import psutil

from bleachbit import IS_WINDOWS
from bleachbit.General import get_real_username, sudo_mode
from bleachbit.Process import (
    _enumerate_ps_aux,
    enumerate_processes,
    is_process_running,
    ProcessInfo,
)
from tests import common


class ProcessTestCase(common.BleachbitTestCase):

    def test_enumerate_processes(self):
        """Test enumerate_processes()"""
        # FIXME: also check the private functions like _enumerate_proc_fs
        processes = list(enumerate_processes())
        self.assertGreater(len(processes), 0)

        for proc in processes:
            self.assertIsInstance(proc, ProcessInfo)
            self.assertIsInstance(proc.pid, int)
            self.assertGreaterEqual(proc.pid, 0, proc)
            self.assertIsInstance(proc.name, str)
            self.assertGreater(len(proc.name), 0)
            self.assertIsInstance(proc.same_user, bool)

    @mock.patch('subprocess.check_output')
    @common.skipIfWindows
    @common.also_with_sudo
    def test_enumerate_ps_aux(self, mock_check_output):
        username = get_real_username()
        ps_out = f"""USER               PID  %CPU %MEM      VSZ    RSS   TT  STAT STARTED      TIME COMMAND
root               703   0.0  0.0  2471428   2792   ??  Ss   20May16   0:01.30 SubmitDiagInfo
alocaluseraccount   681   0.0  0.0  2471568    856   ??  S    20May16   0:00.81 DiskUnmountWatcher
alocaluseraccount   666   0.0  0.0  2507092   3488   ??  S    20May16   0:17.47 SpotlightNetHelper
root               665   0.0  0.0  2497508    512   ??  Ss   20May16   0:11.30 check_afp
alocaluseraccount   646   0.0  0.1  2502484   5656   ??  S    20May16   0:03.62 DataDetectorsDynamicData
alocaluseraccount   632   0.0  0.0  2471288    320   ??  S    20May16   0:02.79 mdflagwriter
alocaluseraccount   616   0.0  0.0  2497596    520   ??  S    20May16   0:00.41 familycircled
alocaluseraccount   573   0.0  0.0  3602328   2440   ??  S    20May16   0:39.64 storedownloadd
alocaluseraccount   572   0.0  0.0  2531184   3116   ??  S    20May16   0:02.93 LaterAgent
{username}   561   0.0  0.0  2471492    584   ??  S    20May16   0:00.21 USBAgent
alocaluseraccount   535   0.0  0.0  2496656    524   ??  S    20May16   0:00.33 storelegacy
root               531   0.0  0.0  2501712    588   ??  Ss   20May16   0:02.40 suhelperd
"""
        mock_check_output.return_value = ps_out

        result = list(_enumerate_ps_aux())
        expected = [
            ProcessInfo(703, 'SubmitDiagInfo', False),
            ProcessInfo(681, 'DiskUnmountWatcher', False),
            ProcessInfo(666, 'SpotlightNetHelper', False),
            ProcessInfo(665, 'check_afp', False),
            ProcessInfo(646, 'DataDetectorsDynamicData', False),
            ProcessInfo(632, 'mdflagwriter', False),
            ProcessInfo(616, 'familycircled', False),
            ProcessInfo(573, 'storedownloadd', False),
            ProcessInfo(572, 'LaterAgent', False),
            ProcessInfo(561, 'USBAgent', True),
            ProcessInfo(535, 'storelegacy', False),
            ProcessInfo(531, 'suhelperd', False),
        ]
        self.assertEqual(result, expected)

        mock_check_output.return_value = 'invalid-input'
        with self.assertRaises(RuntimeError):
            list(_enumerate_ps_aux())

    @common.also_with_sudo
    def test_is_process_running_self(self):
        """Test that this process is running"""
        exe_names = [psutil.Process().name()]
        parent = psutil.Process().parent()
        if parent:
            exe_names.append(parent.name())

        if IS_WINDOWS:
            exe_names.extend([e.swapcase() for e in exe_names])

        # Under sudo the test process (and its parent) run as root, while
        # 'same_user' intentionally compares against the real (non-root)
        # user, so the current process is not considered 'same_user'.
        # The require_same_user=True assertion is only valid when the
        # current process actually runs as the real user.
        check_same_user = not sudo_mode()

        for exe_name in exe_names:
            self.assertTrue(is_process_running(exe_name, False))
            if check_same_user:
                self.assertTrue(is_process_running(exe_name, True))

    @common.also_with_sudo
    def test_is_process_running_does_not_exist(self):
        """Test is_process_running() with processes that do not exist"""
        this_proc_name = psutil.Process().name()
        tests = [
            'does_not_exist',
            '',
            '*',
            this_proc_name + 'x',
            'x' + this_proc_name,
            this_proc_name[1:]
        ]
        # if current process is 'python3', then not unlikely 'python' could
        # be running too, so omit: this test `psutil.Process().name()[:-1]`
        if not this_proc_name == 'python3':
            tests.append(this_proc_name[:-1])
        if not IS_WINDOWS:
            tests.append(this_proc_name.swapcase())
        for exename in tests:
            for require_same_user in (False, True):
                with self.subTest(exename=exename, require_same_user=require_same_user):
                    self.assertFalse(is_process_running(exename, require_same_user),
                                     f'is_running({exename}, {require_same_user})')

    @common.skipIfWindows
    @common.also_with_sudo
    def test_is_process_running_linux(self):
        """Unit test for method is_process_running()"""
        # Do not use get_executable() here because of how it
        # handles symlinks.
        # sys.executable may look like /usr/bin/python3
        # When running under `env -i`, then sys.executable is empty.
        if sys.executable:
            exe = os.path.basename(os.path.realpath(sys.executable))
        else:
            try:
                with open('/proc/self/stat', 'r', encoding='utf-8') as f:
                    exe = f.read().split()[1].strip('()')
            except (IOError, IndexError):
                self.skipTest(
                    "Could not determine current process name from /proc/self/stat")
        tests = [
            # (expected, exe, require_same_user)
            (True, exe, False),  # Check the actual process name
            (True, exe, True),  # Check the actual process name
        ]
        # These processes may be running but not by the current user.
        non_user_exes = ('polkitd', 'bluetoothd', 'NetworkManager',
                         'gdm3', 'snapd', 'systemd-journald')
        tests += [(False, name, True) for name in non_user_exes]
        # These do not exist.
        tests += [
            (False, 'does-not-exist', True),
            (False, 'does-not-exist', False),
        ]
        for expected, exename, require_same_user in tests:
            with self.subTest(exename=exename, require_same_user=require_same_user):
                self.assertEqual(is_process_running(exename, require_same_user), expected,
                                 f'is_running({exename}, {require_same_user})')

    def test_missing_psutil(self):
        """Process should be importable without psutil and set _has_psutil=False."""
        with common.mock_missing_package(
                'psutil',
                clear_prefixes=('bleachbit.Process',)):
            import bleachbit.Process as Process
            self.assertFalse(Process._has_psutil)

    @common.skipUnlessWindows
    def test_is_process_running_windows(self):
        """Test is_process_running() on Windows"""
        # winlogon.exe runs on Windows XP and Windows 7
        # explorer.exe did not run Appveyor, but it does as of 2025-01-25.
        # svchost.exe runs both as system and current user on Windows 11
        # svchost.exe does not run as same user on AppVeyor and Windows Server 2012.
        tests = ((True, 'winlogon.exe', False),
                 (True, 'WinLogOn.exe', False),
                 (False, 'doesnotexist.exe', False),
                 (True, 'explorer.exe', True),
                 (True, 'svchost.exe', False),
                 (True, 'services.exe', False),
                 (False, 'services.exe', True),
                 (True, 'wininit.exe', False),
                 (False, 'wininit.exe', True))

        for expected, exename, require_same_user in tests:
            with self.subTest(exename=exename, require_same_user=require_same_user):
                result = is_process_running(exename, require_same_user)
                self.assertEqual(
                    expected, result, f'Expecting is_process_running({exename}, {require_same_user}) = {expected}, got {result}')

    def test_missing_psutil_import(self):
        """Process should be importable without psutil and set _has_psutil=False."""
        with common.mock_missing_package(
                'psutil',
                clear_prefixes=('bleachbit.Process',)):
            import bleachbit.Process as Process
            self.assertFalse(Process._has_psutil)
            if IS_WINDOWS:
                with self.assertRaises(RuntimeError):
                    list(Process.enumerate_processes())


#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Utility script to discover and run tests marked for sudo execution.
Tests are marked with the @also_with_sudo decorator.

Usage:
    make tests-with-sudo
"""

import inspect
import os
import shlex
import shutil
import subprocess
import sys
import unittest
from importlib import import_module

# Add parent directory to path for imports
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

# Import common test utilities
from tests import common


def discover_sudo_tests():
    """Discover all test methods marked with @also_with_sudo decorator"""
    sudo_tests = []

    # Import test modules to scan for decorated tests
    test_modules = [
        'tests.TestCLI',
        'tests.TestGeneral',
        'tests.TestUnix',
    ]

    for module_name in test_modules:
        try:
            module = import_module(module_name)

            # Find all test classes in the module
            for name, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and issubclass(obj, unittest.TestCase):
                    # Find all test methods in the class
                    for method_name, method in inspect.getmembers(obj):
                        if (method_name.startswith('test_') and
                                hasattr(method, '_also_with_sudo')):
                            test_path = f"{module_name}.{name}.{method_name}"
                            sudo_tests.append(test_path)

        except ImportError as e:
            print(f"Warning: Could not import {module_name}: {e}")

    return sudo_tests


def _build_test_command(test_list):
    """Return command list to run unittest (optionally under coverage)."""
    coverage_runner = os.environ.get('BLEACHBIT_COVERAGE_RUNNER') or ''
    coverage_file = os.environ.get('BLEACHBIT_COVERAGE_FILE') or ''

    if coverage_runner.strip():
        runner_parts = shlex.split(coverage_runner)
        if runner_parts and not os.path.isabs(runner_parts[0]):
            absolute = shutil.which(runner_parts[0])
            if absolute:
                runner_parts[0] = absolute
    else:
        runner_parts = [sys.executable]

    return runner_parts + ['-m', 'unittest'] + test_list + ['-v'], coverage_file


def run_sudo_tests(test_list):
    """Run a list of tests with sudo privileges"""
    if not test_list:
        print("No tests marked with @requires_sudo found")
        return True

    print(f"Found {len(test_list)} tests marked for sudo execution:")
    for test in test_list:
        print(f"  - {test}")

    print(f"\n{'=' * 60}")
    print("Running sudo tests...")
    print('=' * 60)

    base_cmd, coverage_file = _build_test_command(test_list)

    if os.name == 'posix' and os.geteuid() == 0:
        sudo_uid = os.getenv('SUDO_UID')
        if sudo_uid and sudo_uid != '0':
            print("\nDetected EUID=0 with SUDO_UID preserved; skipping nested sudo.")
            cmd = base_cmd
        else:
            print("\nERROR: Do not run tests/test_with_sudo.py with sudo or as root.")
            print("Run it unprivileged; it will invoke sudo internally.")
            print("Try: make tests-with-sudo or python3 tests/test_with_sudo.py")
            return False
    else:
        cmd = ['sudo', '-E'] + base_cmd

    try:
        with common.set_temporary_env('COVERAGE_FILE', coverage_file or None):
            result = subprocess.run(cmd, check=False)
            return result.returncode == 0
    except Exception as e:
        print(f"Error running sudo tests: {e}")
        return False


def main():
    """Discover and run all tests marked with @also_with_sudo"""
    sudo_tests = discover_sudo_tests()

    if not sudo_tests:
        print("No tests marked with @also_with_sudo found")
        sys.exit(0)

    if run_sudo_tests(sudo_tests):
        print("\nAll tests with sudo passed")
        sys.exit(0)
    else:
        print("\nSome tests with sudo failed")
        sys.exit(1)


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Run all test suites

Usage:
    python3 -m tests.TestAll
"""

import os
import unittest
import sys
import tempfile
import time


try:
    from tests import common
except ImportError as e:
    print(f'Error importing tests: {e}')
    print('try python3 -m tests.TestAll')
    sys.exit(1)


def main():
    """Run all tests"""
    print("""You should use the unittest discovery, it's much nicer:
    python -m unittest discover -p Test*.py                       # run all tests
    python -m unittest tests.TestCLI                              # run only the CLI tests
    python -m unittest tests.TestCLI.CLITestCase.test_encoding    # run only a single test

    Or use pytest for more features:
    pytest -n auto tests/                                         # run all tests in parallel
    pytest tests/TestCLI.py                                       # run only the CLI tests
    pytest tests/TestCLI.py::CLITestCase::test_encoding           # run only a single test""")

    start_time = time.time()
    with tempfile.TemporaryDirectory(prefix='TestAll ' + __name__) as testdir:
        with common.set_temporary_env('BLEACHBIT_TEST_OPTIONS_DIR', testdir):
            suite = unittest.defaultTestLoader.discover(
                os.getcwd(), pattern='Test*.py')
            success = unittest.TextTestRunner(verbosity=2).run(suite).wasSuccessful()

    elapsed_time = time.time() - start_time
    minutes = int(elapsed_time // 60)
    seconds = elapsed_time % 60
    print(f"\nTotal test time: {minutes} minutes {seconds:.2f} seconds")
    sys.exit(success is False)


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test case for RecognizeCleanerML
"""

import os
from unittest import mock

import bleachbit
from tests import common
from bleachbit.Options import options, path_to_option
from bleachbit.RecognizeCleanerML import hashdigest, RecognizeCleanerML


class RecognizeCleanerMLTestCase(common.BleachbitTestCase):
    """Test case for RecognizeCleanerML"""

    def test_hash(self):
        """Unit test for hash()"""
        digest = hashdigest('bleachbit')
        self.assertEqual(len(digest), 128)
        self.assertEqual(digest[1:10], '6382c203e')

    def test_init_without_salt(self):
        """Unit test for __init__ when hashsalt is not set in options"""
        self.assertFalse(options.has_option('hashsalt'))
        # Create a CleanerML file so list_cleanerml_files() returns a file,
        # which exercises __recognized().
        cleaner_dir = bleachbit.personal_cleaners_dir
        os.makedirs(cleaner_dir, exist_ok=True)
        cleaner_file = os.path.join(cleaner_dir, 'test_recognize.xml')
        try:
            with open(cleaner_file, 'w', encoding='utf-8') as f:
                f.write(
                    '<?xml version="1.0"?><cleaner id="test"><label>Test</label></cleaner>')
            with mock.patch('bleachbit.RecognizeCleanerML.cleaner_change_dialog'):
                rcm = RecognizeCleanerML(parent_window=None)
            self.assertIsNotNone(rcm.salt)
            self.assertIsInstance(rcm.salt, str)
            self.assertEqual(len(rcm.salt), 128)
            # Verify __recognized() was exercised by checking the hash was stored.
            self.assertTrue(options.has_option(
                path_to_option(cleaner_file), 'hashpath'))
        finally:
            if os.path.exists(cleaner_file):
                os.remove(cleaner_file)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for the context menu command
"""


import os
import sys
import subprocess
import time
from unittest import mock

import psutil

import bleachbit
from bleachbit.GtkShim import HAVE_GTK
from bleachbit.Options import options
from tests import common

if HAVE_GTK:
    from bleachbit.GuiApplication import Bleachbit


START_TIMEOUT_SECONDS = 20
MAX_PATH_DISPLAY = 200


def is_bleachbit_running_process(process_name, cmdline):
    """Return whether process_name/cmdline represent a running BleachBit GUI/context process."""
    assert isinstance(process_name, str)
    assert isinstance(cmdline, list)
    name_lower = (process_name or '').lower()
    normalized_cmdline = [str(arg).lower() for arg in (cmdline or [])]
    cmdline_str = ' '.join(normalized_cmdline)

    image_is_python = name_lower.startswith('python')
    # With subprocess shell=True, process image is cmd.exe instead of python.
    image_is_cmd = name_lower in ('cmd.exe', 'cmd')
    launches_python = 'python.exe' in cmdline_str or 'python3' in cmdline_str
    if not (image_is_python or (image_is_cmd and launches_python)):
        return False

    has_bleachbit_entrypoint = any(
        'bleachbit.py' in arg
        for arg in normalized_cmdline
    )
    has_context_arg = any(
        '--context-menu' in arg or '--gui' in arg for arg in normalized_cmdline)
    return has_bleachbit_entrypoint and has_context_arg


def wait_for_process_tree_windows(process, timeout=60, poll_interval=0.1):
    """Wait for a process and its grandchildren to complete on Windows.

    This function handles the case where a child process spawns a grandchild
    and then exits. It monitors the process tree to ensure all descendants
    have completed before returning.

    Args:
        process: subprocess.Popen object for the child process
        timeout: Maximum time to wait in seconds (default: 60)
        poll_interval: How often to check process status in seconds (default: 0.1)

    Returns:
        The return code of the child process

    Raises:
        subprocess.TimeoutExpired: If the process tree doesn't complete within timeout
    """

    start_time = time.time()
    child_pid = process.pid
    grandchild_pids = set()
    child_exited = False

    while True:
        elapsed = time.time() - start_time
        if elapsed > timeout:
            process.kill()
            process.wait()
            raise subprocess.TimeoutExpired(process.args, timeout)

        # Add absolute safety timeout to prevent hanging
        if elapsed > timeout + 30:  # Extra 30 seconds safety margin
            # Force kill everything and exit
            try:
                process.kill()
                process.wait()
                # Try to kill any remaining grandchildren
                for pid in grandchild_pids:
                    try:
                        if psutil.pid_exists(pid):
                            proc = psutil.Process(pid)
                            proc.kill()
                    except (psutil.NoSuchProcess, psutil.AccessDenied):
                        pass
            except:
                pass
            raise subprocess.TimeoutExpired(process.args, elapsed)

        if not child_exited:
            child_status = process.poll()
            if child_status is not None:
                child_exited = True
            else:
                try:
                    parent_process = psutil.Process(child_pid)
                    current_children = {
                        child.pid for child in parent_process.children(recursive=True)}
                    grandchild_pids.update(current_children)
                except (psutil.NoSuchProcess, psutil.AccessDenied):
                    pass

        if child_exited:
            if not grandchild_pids:
                return process.returncode

            still_running = set()
            for pid in grandchild_pids:
                try:
                    if psutil.pid_exists(pid):
                        proc = psutil.Process(pid)
                        if proc.is_running() and proc.status() != psutil.STATUS_ZOMBIE:
                            still_running.add(pid)
                except (psutil.NoSuchProcess, psutil.AccessDenied):
                    pass

            grandchild_pids = still_running
            if not grandchild_pids:
                return process.returncode

        time.sleep(poll_interval)


class ApplicationRunningTracker:
    """Tracks processes and window titles during application run detection"""

    def __init__(self, check_window_title=False, timeout=0):
        self.check_window_title = check_window_title
        self.timeout = timeout
        self.seen_processes = set()
        self.seen_window_titles = set()
        self.elapsed_time = None
        # This search is more broader than criteria for whether application
        # is running, in case the other criteria is too strict.
        self.candidate_search = ['bleach', 'python']

    def _check_application_state(self):
        """Check current application state without timeout logic.

        Returns tuple (is_running, window_open) where:
        - is_running: True if bleachbit process found
        - window_open: True if bleachbit window found (Windows only)
        """
        is_process_running = False
        window_open = False

        for process in psutil.process_iter(['name', 'cmdline']):
            try:
                name = process.info['name'] or ''
                cmdline = process.info['cmdline'] or []
                name_and_cmdline = f"{name}: {' '.join(cmdline)[:MAX_PATH_DISPLAY]}"

                # Track all interesting processes
                for candidate in self.candidate_search:
                    if candidate in name_and_cmdline.lower():
                        self.seen_processes.add(name_and_cmdline)

                if is_bleachbit_running_process(name, cmdline):
                    is_process_running = True
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                pass

        # Check window titles on Windows
        if os.name == 'nt' and self.check_window_title:
            opened_windows_titles = common.get_opened_windows_titles()
            window_open = any(
                'BleachBit' == window_title for window_title in opened_windows_titles)

            for window_title in opened_windows_titles:
                for candidate in self.candidate_search:
                    if candidate in window_title.lower():
                        self.seen_window_titles.add(window_title)
                        break

        return is_process_running, window_open

    def _is_running_based_on_checks(self, is_process_running, window_open):
        """Determine if application is running based on check_window_title setting."""
        if self.check_window_title and os.name == 'nt':
            return is_process_running and window_open
        return is_process_running or window_open

    def is_application_running(self, expect_running=True):
        """Check if the application is running with optional timeout polling.

        Args:
            expect_running: If True, poll until timeout waiting for app to start.
                           If False, return immediately.

        Returns:
            bool: True if application is running, False otherwise
        """
        if expect_running and self.timeout > 0:
            start_time = time.time()
            while True:
                is_process_running, window_open = self._check_application_state()
                if self._is_running_based_on_checks(is_process_running, window_open):
                    return True
                self.elapsed_time = time.time() - start_time
                if self.elapsed_time >= self.timeout:
                    return False
                # Add safety check to prevent infinite loops
                if self.elapsed_time > self.timeout + 5:  # Extra safety margin
                    return False
                time.sleep(0.1)
        else:
            is_process_running, window_open = self._check_application_state()
            return self._is_running_based_on_checks(is_process_running, window_open)

    def not_running_msg(self):
        """Returns a string with diagnostic information

        Returns
        - count of window titles
        - list of potential matches for window titles (including those seen earlier)
        - count of processes running
        - list of potential matches for processes (including those seen earlier)
        - check_window_title
        - timeout limit
        - time elapsed waiting
        """
        window_title_str = ""
        if os.name == 'nt' and self.check_window_title:
            opened_window_titles = common.get_opened_windows_titles()
            window_title_count = len(opened_window_titles)
            interesting_window_titles = []

            # Include currently seen interesting titles
            for window_title in opened_window_titles:
                for candidate in self.candidate_search:
                    if candidate in window_title.lower():
                        interesting_window_titles.append(window_title)
                        break

            # Also include titles we saw earlier but don't see now
            missing_titles = self.seen_window_titles - \
                set(opened_window_titles)
            if missing_titles:
                interesting_window_titles.extend(
                    [f"[PREVIOUSLY SEEN] {title}" for title in missing_titles])

            window_title_str = f"currently {window_title_count} window titles, interesting:\n" + \
                "\n".join(
                    f"  - {title}" for title in interesting_window_titles)

        interesting_processes = []
        process_count = 0
        current_processes = set()

        for process in psutil.process_iter(['name', 'cmdline']):
            process_count += 1
            name = process.info['name'] or ''
            cmdline = process.info['cmdline'] or []
            name_and_cmdline = f"{name}: {' '.join(cmdline)[:MAX_PATH_DISPLAY]}"
            current_processes.add(name_and_cmdline)

            for candidate in self.candidate_search:
                if candidate in name_and_cmdline.lower():
                    interesting_processes.append(name_and_cmdline)
                    break

        # Also include processes we saw earlier but don't see now
        missing_processes = self.seen_processes - current_processes
        if missing_processes:
            interesting_processes.extend(
                [f"[PREVIOUSLY SEEN] {proc}" for proc in missing_processes])

        process_str = f"currently {process_count} processes, interesting:\n" + \
            "\n".join(f"  - {proc}" for proc in interesting_processes)

        return f"{window_title_str}\n{process_str}" \
            f"\ncheck_window_title: {self.check_window_title}" \
            f"\ntimeout: {self.timeout}" \
            f"\ntime elapsed: {round(self.elapsed_time, 2)}"


class ExternalCommandTestCase(common.BleachbitTestCase):
    """Test case for the context menu command"""

    @classmethod
    def setUpClass(cls):
        """Set up the test case"""
        cls.old_language = common.get_env('LANGUAGE')
        cls._lang_env = common.set_temporary_env('LANGUAGE', 'en')
        cls._lang_env.__enter__()
        super(ExternalCommandTestCase, ExternalCommandTestCase).setUpClass()
        cls._test_options_env = None
        # This should not be needed because of using CLI arg --no-delete-confirmation.
        options.set('delete_confirmation', False)
        options.commit()
        if 'BLEACHBIT_TEST_OPTIONS_DIR' not in os.environ:
            # Set environment variable for child process.
            cls._test_options_env = common.set_temporary_env(
                'BLEACHBIT_TEST_OPTIONS_DIR', cls.tempdir)
            cls._test_options_env.__enter__()

    @classmethod
    def tearDownClass(cls):
        """Tear down the test case"""
        if cls._test_options_env:
            cls._test_options_env.__exit__(None, None, None)
        cls._lang_env.__exit__(None, None, None)
        super(ExternalCommandTestCase, ExternalCommandTestCase).tearDownClass()

    def assertRunning(self, expect_running=True, check_window_title=True, timeout=START_TIMEOUT_SECONDS):
        """Assert whether BleachBit GUI processes are running or not

        Args:
            expect_running: If True (default), assert that processes ARE running.
                           If False, assert that processes are NOT running.
            check_window_title: Whether to also check for window title (Windows only)
            timeout: Maximum time to wait for process to be running (only used when expect_running=True)
        """
        tracker = ApplicationRunningTracker(
            check_window_title=check_window_title,
            timeout=timeout if expect_running else 0
        )
        is_running = tracker.is_application_running(
            expect_running=expect_running)

        if expect_running and not is_running:
            self.fail(
                f"Expected BleachBit GUI process to be running, but none found\n{tracker.not_running_msg()}")
        elif not expect_running and is_running:
            self.fail("BleachBit GUI process is still running")

    def _context_helper(self, fn_prefix, allow_opened_window=False):
        """Unit test for 'Shred with BleachBit' Windows Explorer context menu command

        Called from
        - test_windows_explorer_context_menu_natural
        - test_context_menu_command_while_the_app_is_running

        This test is more likely an integration test as it imitates user behavior.
        It could be interpreted as TestCLI->test_gui_no-uac_shred_exit
        but it is more explicit to have it here as a separate case.
        It covers a single case where one file is shred without delete confirmation dialog.
        """

        file_to_shred = self.mkstemp(prefix=fn_prefix)
        self.assertExists(file_to_shred)
        if not allow_opened_window:
            self.assertRunning(False)
        shred_command_string = self._get_shred_command_string(file_to_shred)
        self._run_shred_command(shred_command_string,
                                cwd=bleachbit.bleachbit_exe_path)
        self.assertNotExists(file_to_shred)

        if not allow_opened_window:
            self.assertRunning(False)

    def _get_shred_command_string(self, file_to_shred):
        """Build the command string like the context menu command"""
        shred_command_string = (
            # Disable confirmation for automated tests.
            f'{sys.executable} bleachbit.py --no-delete-confirmation --no-first-start '
            f'--context-menu "{file_to_shred}"'  # The pathname must be last.
        )
        return shred_command_string

    def _run_shred_command(self, shred_command_string, cwd=None):
        """Run the shred command

        In case of privilege escalation, the child process
        will launch an elevated process, so the child will exit.
        We must wait for the grandchild to finish.
        """
        process = subprocess.Popen(shred_command_string,
                                   shell=True, cwd=cwd)
        try:
            # It may close too quickly to detect the window title.
            self.assertRunning(check_window_title=False)
            returncode = wait_for_process_tree_windows(process, timeout=60)
        except subprocess.TimeoutExpired:
            process.kill()
            process.wait()
            self.fail("Shred command timed out")
        finally:
            # Ensure process is cleaned up to avoid ResourceWarning
            if process.poll() is None:
                process.kill()
                process.wait()
        self.assertEqual(
            returncode, 0, f"Shred command failed with return code {returncode}")

    def test_windows_explorer_context_menu_natural(self):
        """Test shredding a file with the context menu command in a natural way"""
        # First, test the simple case of a pathname without embedded spaces.
        # Then, test a pathname with a space.
        self.assertRunning(False)
        if not HAVE_GTK:
            self.skipTest('GTK is not available')
        for fn_prefix in ('shred_target_no_spaces', 'shred target with spaces'):
            self._context_helper(fn_prefix)
        self.assertRunning(False)

    @common.skipUnlessWindows
    def test_windows_explorer_context_menu_privilege_escalation(self):
        """Test shredding a file with the context menu command with privilege escalation"""
        if bleachbit.Windows.path_on_network(bleachbit.bleachbit_exe_path):
            self.skipTest('Escalation is not attempted on network paths.')
        if not HAVE_GTK:
            self.skipTest('GTK is not available')
        self.assertRunning(False)
        for fn_prefix in ('shred_target_no_spaces', 'shred target with spaces'):
            self._test_as_non_admin_with_context_menu_path(fn_prefix)
        self.assertRunning(False)

    @common.skipUnlessWindows
    def test_context_menu_command_while_the_app_is_running(self):
        """Test the context menu command while the application is running"""
        self.assertRunning(False)
        if not HAVE_GTK:
            self.skipTest('GTK is not available')
        # Start an idle process that will keep running.
        # The --no-uac flag prevents launching grandchild process.
        # The --no-load-cleaners makes it faster.
        args = [sys.executable, 'bleachbit.py',
                '--gui', '--no-load-cleaners', '--no-check-online-updates', '--no-uac']
        p = subprocess.Popen(args, shell=False)
        try:
            # Let the first process start up before launching the second process.
            self.assertRunning(check_window_title=True)
            try:
                # Run a second process that will shred a file.
                self._context_helper('while_app_is_running',
                                     allow_opened_window=True)
            finally:
                self.assertRunning()
        finally:
            # Ensure the process is always killed, even if assertions fail
            if p.poll() is None:  # Check if process is still running
                p.kill()
                p.wait()
        self.assertRunning(False)

    def _test_as_non_admin_with_context_menu_path(self, fn_prefix):
        """
        This tests covers elevate_privileges in the case where we pretend that we are not admin.
        """
        self.assertTrue(os.name == 'nt')
        file_to_shred = self.mkstemp(prefix=fn_prefix)
        self.assertExists(file_to_shred)

        original = bleachbit.Windows.shell.ShellExecuteEx

        def shell_execute_synchronous(lpVerb='', lpFile='', lpParameters='', nShow=''):
            """
            We need a synchronous call the ShellExecuteEx, so we can assert after it finishes.
            """
            from win32com.shell import shell, shellcon
            import win32event
            import win32file
            bleachbit.Windows.shell.ShellExecuteEx = original
            rc = shell.ShellExecuteEx(lpVerb=lpVerb,
                                      lpFile=lpFile,
                                      lpParameters=lpParameters,
                                      nShow=nShow,
                                      fMask=shellcon.SEE_MASK_NOCLOSEPROCESS,
                                      )
            hproc = rc['hProcess']
            win32event.WaitForSingleObject(hproc, win32event.INFINITE)
            # Close the process handle to prevent ResourceWarning
            win32file.CloseHandle(hproc)
            return rc

        # We redirect the ShellExecuteEx call like this because if we do it in a mock we enter recursion
        # because we call ShellExecuteEx in shell_execute_synchronous
        bleachbit.Windows.shell.ShellExecuteEx = shell_execute_synchronous
        with mock.patch('bleachbit.Windows.shell.IsUserAnAdmin', return_value=False):
            # Do not actually exit the process during the test.
            with mock.patch('bleachbit.GuiApplication.sys.exit'):
                # Mock the command line arguments to be inherited to the child process.
                custom_argv = [
                    'dummy-arg',
                    '--no-delete-confirmation',
                    '--context-menu',
                    file_to_shred,  # The pathname must be last
                ]
                with mock.patch('bleachbit.Windows.sys.argv', custom_argv):
                    Bleachbit(auto_exit=True, shred_paths=[
                              file_to_shred], uac=True)

        self.assertNotExists(file_to_shred)

    def test_is_bleachbit_running_process(self):
        """Test process-matching helper without depending on live processes."""
        # Path as seen on AppVeyor
        py_dir = r'C:\\projects\\bleachbit\\vcpkg_installed\\x86-windows\\tools\\python3\\'
        py_exe = os.path.join(py_dir, 'python.exe')
        tests = (
            # Test runner.
            (False, 'python.exe',
                [
                    py_exe,
                    os.path.join(py_dir, 'coverage.exe'),
                    'run',
                    '--include=bleachbit/*',
                    '-m',
                    'tests.TestAll',
                ]
             ),
            # D-Bus session bus.
            (False, 'gdbus.exe',
                [os.path.join(py_dir, 'gdbus.exe'), '_win32_run_session_bus']
             ),
            # This is the cleaning process.
            (True, 'cmd.exe',
                [
                    'C:\\Windows\\system32\\cmd.exe',
                    '/c',
                    py_exe,
                    'bleachbit.py',
                    '--no-delete-confirmation',
                    '--no-first-start',
                    '--context-menu'  # followed by a pathname to delete
                ]
             ),
            # This is the cleaning process as seen on AppVeyor with shell=True where arguments are grouped.
            (True, 'cmd.exe',
                [
                    'C:\\Windows\\system32\\cmd.exe',
                    '/c',
                    f'{py_exe} bleachbit.py --no-delete-confirmation --no-first-start --context-menu C:\\Users\\appveyor\\AppData'
                ]
             ),
            # This is the idle process.
            (True, 'python.exe',
                [py_exe, 'bleachbit.py', '--gui', '--no-load-cleaners',
                    '--no-check-online-updates', '--no-uac']
             ),
        )
        for expected, process_name, cmdline in tests:
            self.assertEqual(
                expected,
                is_bleachbit_running_process(process_name, cmdline)
            )

    def test_wait_for_grandchild_process(self):
        """Test that wait_for_process_tree_windows waits for grandchild processes"""
        marker_file = self.mkstemp(prefix='grandchild_marker_')
        os.remove(marker_file)
        self.assertNotExists(marker_file)

        grandchild_script = os.path.join(self.tempdir, 'grandchild.py')
        with open(grandchild_script, 'w', encoding='utf-8') as f:
            f.write(f'''import time
import sys
time.sleep(2)
with open({repr(marker_file)}, "w") as f:
    f.write("grandchild completed")
sys.exit(0)
''')

        child_script = os.path.join(self.tempdir, 'child.py')
        with open(child_script, 'w', encoding='utf-8') as f:
            f.write(f'''import subprocess
import sys
import time
p = subprocess.Popen([sys.executable, {repr(grandchild_script)}])
time.sleep(0.5)
p.wait()
sys.exit(0)
''')

        process = subprocess.Popen([sys.executable, child_script])

        start_time = time.time()
        returncode = wait_for_process_tree_windows(process, timeout=10)
        elapsed = time.time() - start_time

        self.assertEqual(returncode, 0)
        self.assertExists(marker_file)
        self.assertGreater(elapsed, 1.5)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test cases for __init__
"""

import os

from bleachbit import IS_POSIX, IS_WINDOWS, get_share_dirs, get_share_path
from tests import common


class InitTestCase(common.BleachbitTestCase):

    """Test cases for __init__"""

    def test_expanduser(self):
        """Unit test for function expanduser()"""
        # already absolute
        test_input = '/home/user/foo'
        test_output = os.path.expanduser(test_input)
        self.assertEqual(test_input, test_output)

        # tilde not at beginning
        test_input = '/home/user/~'
        test_output = os.path.expanduser(test_input)
        self.assertEqual(test_input, test_output)

        # should be expanded
        if IS_WINDOWS:
            test_inputs = ('~', r'~\ntuser.dat')
        elif IS_POSIX:
            test_inputs = ('~', '~/.profile')
        for test_input in test_inputs:
            test_output = os.path.expanduser(test_input)
            self.assertNotEqual(test_input, test_output)
            self.assertExists(test_output)
            if IS_POSIX:
                self.assertTrue(os.path.samefile(
                    test_output, os.path.expanduser(test_input)))

    def test_get_share_dirs(self):
        """Unit test for get_share_dirs()"""
        got_shared_dirs = get_share_dirs()
        # There must be at least one shared directory.
        self.assertGreater(len(got_shared_dirs), 0)
        existing = [d for d in got_shared_dirs if os.path.isdir(d)]
        # At least one of the shared directories must exist.
        self.assertGreater(len(existing), 0,
                           f'No share dir exists among: {got_shared_dirs}')

    def test_get_share_path(self):
        """Unit test for get_share_path()"""
        for fn in ('app-menu.ui', 'protected_path.xml'):
            self.assertExists(get_share_path(fn))
        self.assertIsNone(get_share_path('nonexistent'))


def suite():
    return unittest.makeSuite(InitTestCase)



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test FileUtilities.wipe_path

Example invocations:

# Run unit test that wipes a disk image. This is safe.
sudo python3 -m unittest -v tests.TestWipePath

# Equivalent to above
sudo python3 -m tests.TestWipePath

# Run unit test with specific filesystem
sudo BB_FS=ext3 python3 -m unittest -v tests.TestWipePath

# Wipe a real device---dangerous with all file system types
sudo python3 -m tests.TestWipePath --destroy /dev/sdb4

# Wipe a real device---dangerous with ext3
sudo python3 -m tests.TestWipePath --destroy /dev/sdb4 --format ext3


"""

import argparse
import logging
import os
import shutil
import stat
import sys
import tempfile
import time
import unittest

from bleachbit.FileUtilities import delete, exe_exists, free_space, listdir
from bleachbit.General import run_external
from bleachbit.Wipe import wipe_path
from tests import common

logger = logging.getLogger('bleachbit')

IMAGE_SIZE_BYTES = 2097152
MAX_WIPES = 3
FS_FORMAT_COMMANDS = {
    'ext3': ('/sbin/mkfs.ext3', '-q', '-F'),
    'ext4': ('/sbin/mkfs.ext4', '-q', '-F'),
    # 'xfs': ('/sbin/mkfs.xfs', '-q', '-f'),
    'ntfs': ('/sbin/mkntfs', '-F'),
    'vfat': ('/sbin/mkfs.fat', '-F', '32'),
}
FS_TYPES = tuple(FS_FORMAT_COMMANDS.keys())


def _format_tool_available(fs_type):
    return _resolve_executable(FS_FORMAT_COMMANDS[fs_type][0]) is not None


def create_disk_image(n_bytes, fs_type):
    """Make blank file and return filename

    The filename prefix includes the mkfs program basename to aid debugging.
    """
    (fd, filename) = tempfile.mkstemp(
        suffix='disk-image', prefix=f'bleachbit-wipe-test-{fs_type}-')
    try:
        os.ftruncate(fd, n_bytes)
    finally:
        os.close(fd)
    return filename


def run_or_die(args):
    """Run command and die if it fails

    In case of success, return stdout.
    """
    (rc, stdout, stderr) = run_external(args)
    assert (rc == 0), f"Command failed with return code {rc}, stderr: {stderr}"
    return stdout


def _resolve_executable(path):
    """Return executable path if available, else None."""
    candidates = []
    if os.path.isabs(path):
        candidates.append(path)
        candidates.append(shutil.which(os.path.basename(path)))
    else:
        candidates.append(shutil.which(path))
    for candidate in candidates:
        if candidate and os.path.exists(candidate) and os.access(candidate, os.X_OK):
            return candidate
    return None


def format_filesystem(filename, fs_type):
    """Format filesystem

    FIXME: some file systems require larger image size than default.
    * btrfs requires 114,294,784 bytes
    * xfs requires 300MB

    Potential improvement later: add option to partially fill a block device.
    mkfs.fat accepts argument as block size (1024B)
    mkfs.ntfs accepts argument as sector size (512B)
    mkfs.ext4 accepts argument as 1024B blocks or with optional suffix like 5G
    """
    base_cmd = FS_FORMAT_COMMANDS[fs_type]
    exe_path = _resolve_executable(base_cmd[0])
    if not exe_path:
        raise FileNotFoundError(base_cmd[0])
    cmd = (exe_path, *base_cmd[1:], filename)
    run_or_die(cmd)


def make_dirty(mountpoint):
    """Make filesystem dirty by writing

    Files have "secret" in their name and "sssshhhh" in their contents.
    """
    create_counter = 0
    write_counter = 0
    contents = 'sssshhhh' * 512
    while True:
        try:
            fn = os.path.join(mountpoint, 'secret' + str(create_counter))
            with open(fn, 'w', encoding='utf-8') as f:
                create_counter += 1
                f.write(contents)
                f.flush()
                write_counter += 1
        except IOError:
            logger.error(
                'error while creating or writing to temporary file #%d', create_counter)
            break
    create_counter_str = f"{create_counter:,}"
    write_counter_str = f"{write_counter:,}"
    logger.debug('created %s files and wrote to %s files',
                 create_counter_str, write_counter_str)


def mount_filesystem(device_or_image, mountpoint):
    """Mount filesystem from either a real device or disk image"""
    if _is_block_device(device_or_image):
        option = 'rw,nosuid,nodev,noexec'
    else:
        option = 'loop'
    args = ['mount', '-o', option, device_or_image, mountpoint]
    run_or_die(args)
    logger.debug('mounted %s at %s', device_or_image, mountpoint)


def unmount_filesystem(mountpoint):
    """Unmount filesystem

    In case of error, retry up to 5 times.
    """
    time.sleep(0.5)  # avoid "in use" error
    args = ['umount', mountpoint]
    attempts = 0
    while True:
        (rc, _, stderr) = run_external(args)
        if stderr:
            logger.error(stderr)
        if 0 == rc:
            break
        attempts += 1
        time.sleep(attempts * 2)
        if attempts > 5:
            raise RuntimeError('cannot umount')


def assert_clean(filename, fs_type, expect_clean):
    """Assert cleanliness or dirtiness of a disk image.

    Semantics:
    - If expect_clean is True: assert no contents markers remain.
    - If expect_clean is False: assert contents markers remain (dirty).
    - If expect_clean is None: skip assertion.

    Filename remnants are logged but not asserted because detectability varies by filesystem.

    Returns count of file contents markers.
    """
    # ASCII strings for covers ext*, FAT, and file contents
    (rc_ascii, out_ascii, _err_ascii) = run_external(['strings', filename])
    if rc_ascii != 0:
        out_ascii = ''

    # UTF-16LE strings for covers NTFS filename storage
    out_le = ''
    if 'ntfs' == fs_type:
        (rc_le, out_le, _err_le) = run_external(['strings', '-el', filename])
        if rc_le != 0:
            out_le = ''

    # Filenames may be stored as ASCII or UTF-16LE depending on filesystem
    filename_count = max(out_ascii.count('secret'), out_le.count('secret'))

    # File contents were written as ASCII text
    contents_count = out_ascii.count('sssshhhh')
    contents_count_str = f"{contents_count:,}"
    filename_count_str = f"{filename_count:,}"

    logger.debug('found %s markers in image contents and %s markers in filenames',
                 contents_count_str, filename_count_str)

    if expect_clean is None:
        pass
    elif expect_clean:
        assert contents_count == 0, (
            f'Contents not wiped: found {contents_count_str} markers; '
            f'filename markers: {filename_count_str}; '
            f'fs: {fs_type}')
    else:
        assert contents_count > 0, (
            'No contents marker found; filesystem not dirty '
            f'(filename markers: {filename_count_str}); '
            f'fs: {fs_type}')
    return contents_count


@common.skipIfWindows
def wipe_helper(fs_type, n_bytes=None, device_path=None, passes=MAX_WIPES, sleep_seconds=5):
    """Test FileUtilities.wipe_path on image or block device

    For device image, pass device_path and set n_bytes to None.
    For image, pass n_bytes and set device_path to None.
    """

    assert isinstance(fs_type, str)
    assert fs_type in FS_TYPES

    is_device = False
    if device_path:
        if _is_block_device(device_path):
            filename = device_path
        else:
            raise ValueError(f'Not a block device: {device_path}')
        _real_device_warning(device_path, sleep_seconds)
        is_device = True
        assert n_bytes is None, 'n_bytes must be None when using a block device'
    else:
        assert isinstance(n_bytes, int)
        assert n_bytes > 0
        filename = create_disk_image(n_bytes, fs_type)
        logger.debug('created disk image %s', filename)

    for exe in ('strings', 'df'):
        if not exe_exists(exe):
            raise RuntimeError(f'{exe} not found')

    # format filesystem
    mountpoint = None
    try:
        format_filesystem(filename, fs_type)

        # mount
        mountpoint = tempfile.mkdtemp(prefix='bleachbit-wipe-mountpoint')
        mount_filesystem(filename, mountpoint)
    except Exception:
        # The exception is normal for non-root.
        if mountpoint and os.path.exists(mountpoint):
            delete(mountpoint)
        if not is_device:
            delete(filename)
        raise

    # baseline free disk space
    logger.debug('df for clean filesystem %s', mountpoint)
    logger.debug(run_external(['df', mountpoint])[1])

    # make dirty
    make_dirty(mountpoint)

    # verify dirtiness (require contents marker; filenames may not be detectable on all FS)
    unmount_filesystem(mountpoint)
    content_markers_dirty = assert_clean(filename, fs_type, expect_clean=False)
    mount_filesystem(filename, mountpoint)

    # standard delete
    logger.info('standard delete')
    delete_counter = 0
    for secretfile in listdir(mountpoint):
        if 'secret' not in secretfile:
            # skip lost+found
            continue
        delete(secretfile, shred=False)
        delete_counter += 1
    logger.debug('deleted %s files', f"{delete_counter:,}")

    # check
    logger.debug('df for empty, dirty filesystem %s', mountpoint)
    logger.debug(run_external(['df', mountpoint])[1])

    # verify dirtiness again (require contents marker; filenames may not be detectable)
    unmount_filesystem(mountpoint)
    assert_clean(filename, fs_type, expect_clean=False)
    mount_filesystem(filename, mountpoint)
    expected_free_space = free_space(mountpoint)

    content_markers_list = []
    # measure effectiveness of multiple wipes
    for i in range(1, passes + 1):
        print('*' * 30)
        print(f'* pass {i} for {fs_type} *')
        print('*' * 30)

        # remount
        if i > 1:
            mount_filesystem(filename, mountpoint)

        # really wipe
        logger.info('wiping %s', mountpoint)

        old_level = logging.getLogger('bleachbit.FileUtilities').level
        logging.getLogger('bleachbit.FileUtilities').setLevel(logging.INFO)
        try:
            for _w in wipe_path(mountpoint):
                pass
        finally:
            logging.getLogger('bleachbit.FileUtilities').setLevel(old_level)

        # verify cleaning process freed all space it allocated
        actual_free_space = free_space(mountpoint)
        if not expected_free_space == actual_free_space:
            logger.error('expecting %d free space but got %d',
                         expected_free_space, actual_free_space)

        # unmount
        unmount_filesystem(mountpoint)

        # Require contents to be wiped. Wiping filenames is not supported.
        content_markers = assert_clean(filename, fs_type, None)
        content_markers_list.append(content_markers)
        effectiveness = 100 * (content_markers_dirty -
                               content_markers) / content_markers_dirty
        print(
            f"Pass {i}: Content markers: {content_markers:,}, Effectiveness: {effectiveness:.1f}%")
        if content_markers == 0:
            break

    # show content marker counts to help user understand wiping effectiveness
    if content_markers_list == [0]:
        print('100% effective on first pass!')
    else:
        print(f"\nContent marker counts by pass: {content_markers_list}")
        if len(content_markers_list) > 1:
            if content_markers_list[0] > min(content_markers_list):
                print('Wiping had incremental improvement after first pass')
            else:
                print('Wiping did not improve after first pass')

    if not is_device:
        delete(filename)
    delete(mountpoint)
    if content_markers > 0:
        raise AssertionError(f'Contents not wiped for {fs_type}')


class WipeTestCase(common.BleachbitTestCase):
    """Test case for file wiping operations"""

    @common.skipIfWindows
    @common.also_with_sudo
    def test_wipe(self):
        """Test wiping on several kinds of file systems"""
        if not exe_exists('strings'):
            raise RuntimeError(
                'strings not found: run `sudo apt install binutils` or equivalent')
        requested_fs = FS_TYPES
        # Optional filter example: sudo BB_FS=ext3 python3 -m unittest tests.TestWipePath
        env_fs = os.environ.get('BB_FS') or os.environ.get('bb_fs')
        if env_fs:
            if env_fs not in FS_TYPES:
                raise ValueError(
                    f'BB_FS={env_fs} did not match valid file system type')
            requested_fs = (env_fs,)

        available_fs = tuple(
            fs for fs in requested_fs if _format_tool_available(fs))
        missing_fs = tuple(fs for fs in requested_fs if fs not in available_fs)

        if missing_fs:
            logger.warning('Skipping wipe tests for filesystems lacking mkfs utilities: %s',
                           ', '.join(missing_fs))
        if not available_fs:
            missing_list = ', '.join(missing_fs) or 'unknown'
            self.skipTest(
                f'Missing mkfs utilities for requested filesystems: {missing_list}')

        for fs_type in available_fs:
            if common.have_root():
                wipe_helper(fs_type=fs_type,
                            n_bytes=IMAGE_SIZE_BYTES)
            else:
                # Mount will fail
                with self.assertRaises(AssertionError):
                    wipe_helper(fs_type=fs_type,
                                n_bytes=IMAGE_SIZE_BYTES)


def suite():
    """Return a test suite"""
    return unittest.makeSuite(WipeTestCase)


def _real_device_warning(target_desc, sleep_seconds):
    border = '=' * 78
    msg = (
        f"\n{border}\n"
        f"DANGEROUS OPERATION REQUESTED\n\n"
        f"You requested wiping all data on: {target_desc}\n"
        f"This is DESTRUCTIVE and will lead to DATA LOSS.\n"
        f"Proceeding in {sleep_seconds} seconds. Press Ctrl+C to abort.\n"
        f"{border}\n"
    )
    print(msg)
    for i in range(sleep_seconds, 0, -1):
        print(f"Starting in {i}...", end='\r', flush=True)
        time.sleep(1)
    print(' ' * 40, end='\r', flush=True)


def _is_block_device(path):
    """Return True if path is a block device."""
    try:
        st = os.stat(path)
        return stat.S_ISBLK(st.st_mode)
    except Exception:
        return False


def _ensure_not_mounted(device_path):
    """Exit if the given device is mounted anywhere.

    Uses 'findmnt' to query mount targets for the source device (and its
    realpath). If mounted anywhere, exit safely.
    """
    if not device_path:
        return
    if not exe_exists('findmnt'):
        raise RuntimeError('findmnt not found')
    # Check real path in case of link such as /dev/disk/by-uuid/..
    candidates = [device_path]
    real = os.path.realpath(device_path)
    if real != device_path:
        candidates.append(real)

    for dev in candidates:
        rc, out, _err = run_external(['findmnt', dev])
        if rc == 0 and out.strip():
            print(
                f"Refusing to operate on {device_path}: mounted at {out.strip()}")
            sys.exit(3)


def main():
    """Main"""
    if len(sys.argv) == 1:
        unittest.main()
        return

    parser = argparse.ArgumentParser(
        description=(
            'Run TestWipePath as a destructive tool to wipe free space on a real block device.'
            'By default, without arguments, this script runs unittests.'
            'DANGER! Do not delete your data! Carefully check the options!'
        )
    )
    parser.add_argument('--destroy', required=True,
                        help='Block device (e.g., /dev/sdb1) that will be destroyed!')
    parser.add_argument(
        '--format', help='Comma-separated filesystems to format before wiping (ext3,ext4,ntfs,vfat)')
    parser.add_argument('--passes', type=int, default=1,
                        help='Number of wipe passes (default: 1)')
    parser.add_argument('--sleep', type=int, default=5,
                        help='Seconds to wait before start (default: 5)')
    parser.add_argument('--verbose', action='store_true',
                        help='Enable verbose logging')

    args = parser.parse_args()

    if not common.have_root():
        print('Use sudo to run this script')
        sys.exit(2)

    # Parse formats list if provided
    formats = FS_TYPES
    if args.format:
        raw = [s.strip().lower() for s in args.format.split(',') if s.strip()]
        invalid = [fs for fs in raw if fs not in FS_TYPES]
        if invalid:
            print(f"Error: unsupported filesystem(s): {', '.join(invalid)}")
            sys.exit(2)
        formats = raw if raw else None

    _ensure_not_mounted(args.destroy)

    for fs_type in formats:
        wipe_helper(fs_type=fs_type,
                    device_path=args.destroy,
                    passes=args.passes,
                    sleep_seconds=args.sleep,
                    )
    sys.exit(0)


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for module Cleaner
"""

import glob
import logging
import os
import shutil
from xml.dom.minidom import parseString

import bleachbit
from bleachbit import IS_WINDOWS, IS_POSIX
from bleachbit.Action import ActionProvider, Command
from bleachbit.Cleaner import Cleaner, backends, create_simple_cleaner, register_cleaners
import bleachbit.FileUtilities

from tests import common

logger = logging.getLogger('bleachbit')


def action_to_cleaner(action_str):
    """Given an action XML fragment, return a cleaner"""
    return actions_to_cleaner([action_str])


def actions_to_cleaner(action_strs):
    """Given multiple action XML fragments, return one cleaner"""

    cleaner = Cleaner()
    for count, action_str in enumerate(action_strs, start=1):
        dom = parseString(action_str)
        action_node = dom.childNodes[0]
        command = action_node.getAttribute('command')
        provider = None
        for actionplugin in ActionProvider.plugins:
            if actionplugin.action_key == command:
                provider = actionplugin(action_node)
        cleaner.add_action(f'option{count}', provider)
        cleaner.add_option(
            f'option{count}', f'name{count}', f'description{count}')
    return cleaner


def register_all_cleaners():
    """Register all cleaners for testing

    _patch_options_paths() changes the options directory during testing
    to a clean directory, so by default it will not have Winapp2.ini.
    """
    if IS_WINDOWS:
        from tests.TestWinapp import get_winapp2  # pylint: disable=import-outside-toplevel

        os.makedirs(bleachbit.personal_cleaners_dir, exist_ok=True)
        print(f"personal_cleaners_dir: {bleachbit.personal_cleaners_dir}")
        shutil.copyfile(
            get_winapp2(),
            os.path.join(bleachbit.personal_cleaners_dir, 'winapp2.ini'),
        )
    if not backends:
        list(register_cleaners())
    assert len(backends) > 1


class CleanerTestCase(common.BleachbitTestCase):

    def test_add_action(self):
        """Unit test for Cleaner.add_action()"""
        self.actions = []
        if IS_WINDOWS:
            self.actions += [
                '<action command="delete" search="file" path="$WINDIR\\explorer.exe"/>',
                '<action command="delete" search="glob" path="$WINDIR\\system32\\*.dll"/>',
                '<action command="delete" search="walk.files" path="$WINDIR\\system32\\"/>',
                '<action command="delete" search="walk.all" path="$WINDIR\\system32\\"/>']
        elif IS_POSIX:
            print(__file__)
            self.actions += [
                f'<action command="delete" search="file" path="{__file__}"/>',
                '<action command="delete" search="glob" path="/bin/*sh"/>',
                '<action command="delete" search="walk.files" path="/bin/"/>',
                '<action command="delete" search="walk.all" path="/var/log/"/>']
        else:
            raise AssertionError('Unknown OS.')
        self.assertGreater(len(self.actions), 0)

        for action_str in self.actions:
            cleaner = action_to_cleaner(action_str)
            count = 0
            for cmd in cleaner.get_commands('option1'):
                for result in cmd.execute(False):
                    self.assertEqual(result['n_deleted'], 1)
                    pathname = result['path']
                    self.assertLExists(
                        pathname, f"Does not exist: '{pathname}'")
                    count += 1
                    common.validate_result(self, result)
            self.assertGreater(count, 0, f"No files found for {action_str}")
        # should yield nothing
        cleaner.add_option('option2', 'name2', 'description2')
        for cmd in cleaner.get_commands('option2'):
            print(cmd)
            raise AssertionError('option2 should yield nothing')
        # should fail
        self.assertRaises(
            RuntimeError, cleaner.get_commands('option3').__next__)

    def test_auto_hide(self):
        for key in sorted(backends):
            self.assertIsInstance(backends[key].auto_hide(), bool)

    def test_create_simple_cleaner(self):
        """Unit test for method create_simple_cleaner"""
        dirname = self.mkdtemp(prefix='bleachbit-test-create-simple-cleaner')
        filename1 = os.path.join(dirname, '1')
        common.touch_file(filename1)
        # test Cyrillic for https://bugs.launchpad.net/bleachbit/+bug/1541808
        filename2 = os.path.join(dirname, 'чистый')
        common.touch_file(filename2)
        targets = [filename1, filename2, dirname]
        cleaner = create_simple_cleaner(targets)
        for cmd in cleaner.get_commands('files'):
            # preview
            for result in cmd.execute(False):
                common.validate_result(self, result)
            # delete
            list(cmd.execute(True))

        for target in targets:
            self.assertNotExists(target)

    def test_create_simple_cleaner_recursive(self):
        """Verify create_simple_cleaner deletes recursive directory trees"""
        # Build:  root/
        #           file.txt
        #           sub1/
        #             file.txt
        #             sub2/
        #               file.txt
        root = self.mkdir('csc-recursive')  # relative to self.tempdir
        sub1 = self.mkdir(os.path.join(root, 'sub1'))
        sub2 = self.mkdir(os.path.join(sub1, 'sub2'))
        dirnames = [root, sub1, sub2]
        for dirname in dirnames:
            common.touch_file(os.path.join(dirname, 'file.txt'))

        cleaner = create_simple_cleaner([root])
        for cmd in cleaner.get_commands('files'):
            list(cmd.execute(True))

        for dirname in dirnames:
            self.assertNotExists(dirname)

    def test_get_name(self):
        for key in sorted(backends):
            self.assertIsString(backends[key].get_name())

    def test_get_description(self):
        for key in sorted(backends):
            self.assertIsString(key)
            self.assertIsInstance(backends[key], Cleaner)
            desc = backends[key].get_description()
            if desc is not None:
                self.assertIsString(
                    desc, msg="description for '%s' is '%s'" % (key, desc))

    def test_get_options(self):
        for key in sorted(backends):
            for (test_id, name) in backends[key].get_options():
                self.assertIsString(
                    test_id, msg='%s.%s is not a string' % (key, test_id))
                self.assertIsString(name)

    def test_get_commands(self):
        for key in sorted(backends):
            logger.debug("test_get_commands: key='%s'", key)
            for (option_id, __name) in backends[key].get_options():
                for cmd in backends[key].get_commands(option_id):
                    for result in cmd.execute(really_delete=False):
                        if result != True:
                            break
                        common.validate_result(self, result)
        # make sure trash and tmp don't return the same results
        if IS_WINDOWS:
            return

        def get_files(option_id):
            ret = []
            register_all_cleaners()
            for cmd in backends['system'].get_commands(option_id):
                result = next(cmd.execute(False))
                ret.append(result['path'])
            return ret
        trash_paths = get_files('trash')
        tmp_paths = get_files('tmp')
        for tmp_path in tmp_paths:
            self.assertNotIn(tmp_path, trash_paths)

    def test_no_files_exist(self):
        """Verify only existing files are returned

        It monkeypatches file system functions to return no files.
        Then, it tries every cleaning option.
        """
        _exists = os.path.exists
        _iglob = glob.iglob
        _isdir = os.path.isdir
        _lexists = os.path.lexists
        _listdir = os.listdir
        _oswalk = os.walk
        _fu_walk = bleachbit.FileUtilities.walk
        try:
            glob.iglob = lambda path, *args, **kwargs: []
            os.path.exists = lambda path: False
            os.path.isdir = lambda path: False
            os.path.lexists = lambda path: False
            os.listdir = lambda path: []
            os.walk = lambda top, topdown=True, onerror=None, followlinks=False: []
            bleachbit.FileUtilities.walk = lambda top, topdown=True, onerror=None, followlinks=False: []
            for key in sorted(backends):
                for (option_id, __name) in backends[key].get_options():
                    for cmd in backends[key].get_commands(option_id):
                        # Some cleaners can still return a command, so next run
                        # a preview.
                        for result in cmd.execute(really_delete=False):
                            if not result:
                                break
                            msg = f"Expected no files to be deleted but got '{result}'"
                            self.assertNotIsInstance(cmd, Command.Delete, msg)
        finally:
            glob.iglob = _iglob
            os.path.exists = _exists
            os.path.isdir = _isdir
            os.path.lexists = _lexists
            os.listdir = _listdir
            os.walk = _oswalk
            bleachbit.FileUtilities.walk = _fu_walk

    def test_register_cleaners(self):
        """Unit test for register_cleaners"""
        register_all_cleaners()
        register_all_cleaners()
        backends.clear()
        register_all_cleaners()

    @common.skipIfWindows  # FIXME later: reevaluate
    @common.skipUnlessDestructive
    def test_system_recent_documents(self):
        """Clean recent documents in GTK"""
        from bleachbit.GtkShim import (
            Gtk, Gio, GLib, suppress_pygobject_asyncio_warnings)
        mgr = Gtk.RecentManager().get_default()
        fn = self.mkstemp(suffix='.txt')
        self.assertExists(fn)
        uri = Gio.File.new_for_path(fn).get_uri()
        self.assertTrue(mgr.add_item(uri))
        GLib.idle_add(Gtk.main_quit)
        # PyGObject 3.56.2 calls deprecated asyncio APIs in Gtk.main(),
        # which breaks tests run with warnings as errors.
        with suppress_pygobject_asyncio_warnings():
            Gtk.main()  # process the addition
        GLib.idle_add(Gtk.main_quit)
        self.assertGreater(len(mgr.get_items()), 0)
        self.assertTrue(mgr.has_item(uri))

        register_all_cleaners()
        for cmd in backends['system'].get_commands('recent_documents'):
            for result in cmd.execute(really_delete=True):
                common.validate_result(self, result, True)

        self.assertEqual(len(mgr.get_items()), 0)

    @common.skipIfWindows
    def test_whitelist(self):
        """Unit test for method Cleaner.whitelisted"""
        tests = [
            ('/tmp/.truecrypt_aux_mnt1/control', True),
            ('/tmp/.truecrypt_aux_mnt1/volume', True),
            ('/tmp/.vbox-foo-ipc/lock', True),
            ('/tmp/.wine-500/server-806-102400f/lock', True),
            ('/tmp/gconfd-foo/lock/ior', True),
            ('/tmp/ksocket-foo/Arts_SoundServerV2', True),
            ('/tmp/ksocket-foo/secret-cookie', True),
            ('/tmp/orbit-foo/bonobo-activation-register-a9cd6cc4973af098918b154c4957a93f.lock', True),
            ('/tmp/orbit-foo/bonobo-activation-register.lock', True),
            ('/tmp/orbit-foo/bonobo-activation-server-a9cd6cc4973af098918b154c4957a93f-ior', True),
            ('/tmp/orbit-foo/bonobo-activation-server-ior', True),
            ('/tmp/pulse-foo/pid', True),
            ('/tmp/tmpsDOBFd', False),
            ('~/.cache/obexd', True),
            ('~/.cache/obexd/', True),
            ('~/.cache/obexd/foo', True),
            ('~/.cache/obex', False),
            ('~/.cache/obexd-foo', False),
            ('~/.cache/kwin', True),
            ('~/.cache/kwi', False),
            ('~/.cache/kwi/foo.txt', False),
            ('~/.cache/kwin/test_file', True),
            ('~/.cache/mesa_shader_cache', True),
            ('~/.cache/mesa_shader_cache/test_file', True),
            ('~/.cache/plasmashell', True),
            ('~/.cache/plasmashell/test_file', True),
            ('~/.cache/icon-cache.kcache', True),
            ('~/.cache/plasma_theme_breeze-light_v5.103.0.kcache', True),
            ('~/.cache/drkonqi', True),
            ('~/.cache/drkonqi/test_file', True),
            ('~/.cache/mesa_shader_cache_db', True),
            ('~/.cache/mesa_shader_cache_db/test_file', True),
            ('~/.cache/qtshadercache-x86_64-little_endian-lp64', True),
            ('~/.cache/qtshadercache-x86_64-little_endian-lp64/test_file', True),
            ('~/.cache/plasma_theme_default.kcache', True)
        ]
        register_all_cleaners()
        for pathname, expected in tests:
            path = os.path.expanduser(
                pathname) if pathname.startswith('~') else pathname
            self.assertEqual(
                backends['system'].whitelisted(path), expected, pathname)
        # Make sure directory ~/.cache/obexd is ignored
        # https://github.com/bleachbit/bleachbit/issues/572
        obexd_dir = os.path.expanduser('~/.cache/obexd')
        if not os.path.exists(obexd_dir):
            os.makedirs(obexd_dir)
        obexd_fn = os.path.join(obexd_dir, 'bleachbit-test')
        common.touch_file(obexd_fn)
        for cmd in backends['system'].get_commands('cache'):
            for _result in cmd.execute(really_delete=False):
                self.assertNotEqual(cmd.path, obexd_fn)
                self.assertFalse('/.cache/obexd/' in cmd.path)
        from bleachbit.FileUtilities import delete
        delete(obexd_fn, ignore_missing=True)

    def test_custom(self):
        """Test system.custom"""
        from bleachbit.Options import options
        original_custom_paths = options.get_custom_paths()

        register_all_cleaners()
        test_pathname = os.path.join(self.tempdir, 'foo')

        # Check that custom cleaner doesn't iterate non-existent object
        for obj_type in ('folder', 'file'):
            options.set_custom_paths([(obj_type, test_pathname)])
            for cmd in backends['system'].get_commands('custom'):
                results = list(cmd.execute(really_delete=False))
                self.assertEqual(len(results), 0)

        # Create the file
        common.touch_file(test_pathname)
        self.assertExists(test_pathname)

        # Check it returns the file now
        for cmd in backends['system'].get_commands('custom'):
            results = list(cmd.execute(really_delete=False))
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['path'], test_pathname)

        # Prepare an empty folder.
        test_dirname = os.path.join(self.tempdir, 'subdir')
        os.makedirs(test_dirname)
        self.assertExists(test_dirname)
        options.set_custom_paths([('folder', test_dirname)])

        # Check it returns the folder now.
        for cmd in backends['system'].get_commands('custom'):
            results = list(cmd.execute(really_delete=False))
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['path'], test_dirname)

        # Restore the original settings.
        options.set_custom_paths(original_custom_paths)
